import argparse
import bisect
import copy
import hashlib
import json
import multiprocessing
import os
from collections import defaultdict
from contextlib import ExitStack
from copy import deepcopy
from dataclasses import dataclass, field
from itertools import chain
from queue import Queue
from tempfile import TemporaryDirectory
from typing import Any, Dict, Generator, List, Tuple, Type, TypeVar, Union

import blingfire
import msgspec
import numpy as np
import smart_open
import tldextract
import tqdm

from dolma.core.data_types import InputSpec, OutputSpec
from dolma.core.parallel import BaseParallelProcessor
from dolma.core.paths import glob_path

T = TypeVar("T", bound=Type["BaseStatsProcessor"])

PRONOUNS = (
    ("she", "her", "her", "hers", "herself"),
    ("he", "him", "his", "his", "himself"),
    ("they", "them", "their", "theirs", "themselves"),
    ("ze", "hir", "hir", "hirs", "hirself"),
    ("ze", "zir", "zir", "zirs", "zirself"),
    ("xey", "xem", "xyr", "xyrs", "xemself"),
    ("ae", "aer", "aer", "aers", "aerself"),
    ("e", "em", "eir", "eirs", "emself"),
    ("ey", "em", "eir", "eirs", "eirself"),
    ("fae", "faer", "faer", "faers", "faerself"),
    ("fey", "fem", "feir", "feirs", "feirself"),
    ("hu", "hum", "hus", "hus", "humself"),
    ("it", "it", "its", "its", "itself"),
    ("jee", "jem", "jeir", "jeirs", "jemself"),
    ("kit", "kit", "kits", "kits", "kitself"),
    ("ne", "nem", "nir", "nirs", "nemself"),
    ("peh", "pehm", "peh's", "peh's", "pehself"),
    ("per", "per", "per", "pers", "perself"),
    ("sie", "hir", "hir", "hirs", "hirself"),
    ("se", "sim", "ser", "sers", "serself"),
    ("shi", "hir", "hir", "hirs", "hirself"),
    ("si", "hyr", "hyr", "hyrs", "hyrself"),
    ("they", "them", "their", "theirs", "themself"),
    ("thon", "thon", "thons", "thons", "thonself"),
    ("ve", "ver", "vis", "vis", "verself"),
    ("ve", "vem", "vir", "virs", "vemself"),
    ("vi", "ver", "ver", "vers", "verself"),
    ("vi", "vim", "vir", "virs", "vimself"),
    ("vi", "vim", "vim", "vims", "vimself"),
    ("xie", "xer", "xer", "xers", "xerself"),
    ("xe", "xem", "xyr", "xyrs", "xemself"),
    ("xey", "xem", "xeir", "xeirs", "xemself"),
    ("yo", "yo", "yos", "yos", "yosself"),
    ("ze", "zem", "zes", "zes", "zirself"),
    ("ze", "mer", "zer", "zers", "zemself"),
    ("zee", "zed", "zeta", "zetas", "zedself"),
    ("zie", "zir", "zir", "zirs", "zirself"),
    ("zie", "zem", "zes", "zes", "zirself"),
    ("zie", "hir", "hir", "hirs", "hirself"),
    ("zme", "zmyr", "zmyr", "zmyrs", "zmyrself"),
)


@dataclass
class Domains:
    pages: Dict[str, int] = field(default_factory=dict)
    words: Dict[str, int] = field(default_factory=dict)
    _size: int = 100_000

    def add(self, domain: str, count_words: int, count_pages: int = 1, no_limit: bool = False) -> bool:
        if domain not in self.pages:
            if self._size < len(self.pages) and not no_limit:
                return False
            self.pages[domain] = 0
            self.words[domain] = 0

        self.pages[domain] += count_pages
        self.words[domain] += count_words
        return True

    def shrink(self, to_size: bool = False) -> bool:
        threshold = 1
        if to_size:
            # find the threshold that will keep the top self._size domains
            prob = max((1 - self._size / len(self.pages)) * 100, 0)
            threshold = max(threshold, round(np.percentile(sorted(self.pages.values()), prob)))

        previous_size = len(self.pages)
        self.pages = {k: v for k, v in self.pages.items() if v > threshold}
        self.words = {k: v for k, v in self.words.items() if k in self.pages}
        current_size = len(self.pages)
        return previous_size < current_size

    def to_dict(self) -> Dict[str, Any]:
        return {
            "pages": {k: v for k, v in sorted(self.pages.items(), key=lambda e: -e[1])},
            "words": {k: v for k, v in sorted(self.words.items(), key=lambda e: -e[1])},
        }

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> "Domains":
        return cls(pages=d["pages"], words=d["words"])

    def merge(self, other: "Domains", inplace: bool = True, shrink: bool = False) -> "Domains":
        to_return = self if inplace else deepcopy(self)

        for page in other.pages:
            to_return.add(domain=page, count_words=other.words[page], count_pages=other.pages[page], no_limit=True)

        if shrink:
            to_return.shrink(to_size=True)

        return to_return


@dataclass
class Counts:
    documents: int = 0
    tokens: int = 0
    domains: Domains = field(default_factory=Domains)
    pronouns: Dict[str, int] = field(default_factory=lambda: {k: 0 for k in chain.from_iterable(PRONOUNS)})
    _flush: int = 250_000
    _current: int = 0

    def shrink(self) -> bool:
        self._current += 1
        if self._current >= self._flush:
            self._current = 0
            self.domains.shrink()
            return True

        return False

    def add(self, text: str, url: str) -> bool:
        if not (text := text.strip()):
            return False

        words = [w.lower() for w in blingfire.text_to_words(text).split()]
        extracted_url = tldextract.extract(url)
        domain = ".".join(extracted_url[1:]).lower()

        for w in words:
            if w in self.pronouns:
                self.pronouns[w] += 1

        self.documents += 1
        self.tokens += len(words)
        self.domains.add(domain=domain, count_words=len(words))

        self.shrink()
        return True

    def to_dict(self) -> Dict[str, Any]:
        return {
            "domains": self.domains.to_dict(),
            "pronouns": {k: v for k, v in sorted(self.pronouns.items(), key=lambda e: -e[1])},
            "documents": self.documents,
            "words": self.tokens,
        }

    @classmethod
    def from_dict(cls, d: Dict[str, Any]) -> "Counts":
        return cls(
            documents=d["documents"],
            tokens=d["words"],
            domains=Domains.from_dict(d["domains"]),
            pronouns=d["pronouns"],
        )

    def merge(self, other: "Counts", inplace: bool = True, shrink: bool = False) -> "Counts":
        (to_return := self if inplace else deepcopy(self)).documents += other.documents
        to_return.tokens += other.tokens
        to_return.domains.merge(other.domains, inplace=True, shrink=shrink)
        for pronoun, count in other.pronouns.items():
            to_return.pronouns[pronoun] += count
        return to_return


class Registry:
    __registry__: Dict[str, Type["BaseStatsProcessor"]] = {}

    @classmethod
    def add(cls, obj: T) -> T:
        cls.__registry__[obj.__name__] = obj
        return obj

    @classmethod
    def get(cls, name: str) -> Type["BaseStatsProcessor"]:
        return cls.__registry__[name]

    @classmethod
    def all(cls) -> Generator[Tuple[str, Type["BaseStatsProcessor"]], None, None]:
        yield from cls.__registry__.items()


class BaseStatsProcessor(BaseParallelProcessor):
    documents: Union[str, List[str]]
    stats: str

    @classmethod
    def increment_progressbar(
        cls, queue: "Queue[Union[Tuple[int, ...], None]]", /, files: int = 0, documents: int = 0
    ) -> Dict[str, int]:
        return super().increment_progressbar(queue, files=files, documents=documents)

    @staticmethod
    # read all paths in using threads
    def _read_json(path: str) -> Counts:
        with smart_open.open(path, "rt") as source_file:
            return msgspec.json.decode(source_file.read())

    @classmethod
    def _merge_dicts(cls, d1, d2):
        d1 = copy.deepcopy(d1)
        for k, v in d2.items():
            if isinstance(v, dict):
                d1[k] = cls._merge_dicts(d1.get(k, {}), v)
            else:
                d1[k] = d1.get(k, 0) + v
        return d1

    @classmethod
    def _run_parallel_processor(cls, stats_root: str, num_workers: int, debug: bool, **process_single_kwargs: Any):
        with TemporaryDirectory() as tempdir:
            h = hashlib.md5()
            for path in cls.documents if isinstance(cls.documents, list) else [cls.documents]:
                h.update(path.encode())

            metadata = os.path.join(tempdir, h.hexdigest())
            processor = cls(
                source_prefix=cls.documents,
                destination_prefix=stats_root,
                metadata_prefix=metadata,
                num_processes=num_workers,
                debug=debug,
            )
            processor(**process_single_kwargs)

    @classmethod
    def cli(cls, num_workers: int = 1, debug: bool = False, **process_single_kwargs: Any) -> None:
        stats_root = cls.stats.split("*", 1)[0].rstrip("/")

        cls._run_parallel_processor(
            stats_root=stats_root,
            num_workers=num_workers,
            debug=debug,
            **process_single_kwargs,
        )

        paths = list(glob_path(cls.stats))
        counts: dict = {}

        with multiprocessing.Pool(num_workers) as pool:
            data = (cls._read_json(path) for path in paths) if debug else pool.imap(cls._read_json, paths)

            for content in tqdm.tqdm(data, desc=f"Merging {cls.__name__} stats", unit=" files", total=len(paths)):
                counts = cls._merge_dicts(counts, content)

        summary_dest = f"{stats_root}/summary.json"
        with smart_open.open(summary_dest, "wt") as destination_file:
            destination_file.write(json.dumps(counts, indent=2, sort_keys=True))


@Registry.add
class books(BaseStatsProcessor):
    documents = "s3://ai2-llm/pretraining-data/sources/olmo-mix/v1/documents/books/*.gz"
    stats = "s3://ai2-llm/stats/olmo-mix/v1/books/gutenberg/*.gz"

    @classmethod
    def process_single(
        cls, source_path: str, destination_path: str, queue: "Queue[Union[Tuple[int, ...], None]]", **kwargs: Any
    ):
        os.environ["TOKENIZERS_PARALLELISM"] = "false"

        # for the data sheet, what statistics you think we should include? I could
        # do # of docs, # tokens, distribution of URLs, pronouns, s2 FOS, stack
        # languages?
        decoder = msgspec.json.Decoder(InputSpec)
        documents = words = 0
        interval = 10_000

        with smart_open.open(source_path, "rb") as source_file:
            for line in source_file:
                document = decoder.decode(line)
                documents += 1
                words += len(blingfire.text_to_words(document.text).split())

                if documents % interval == 0:
                    cls.increment_progressbar(queue, documents=interval)

        cls.increment_progressbar(queue, files=1, documents=documents % interval)

        with smart_open.open(destination_path, "wt") as destination_file:
            destination_file.write(json.dumps({"documents": documents, "words": words}, indent=2))


@Registry.add
class wiki(books):
    documents = "s3://ai2-llm/pretraining-data/sources/olmo-mix/v1/documents/wiki/*.gz"
    stats = "s3://ai2-llm/stats/olmo-mix/v1/wiki/en_simple/*.gz"


@Registry.add
class cc_v1(BaseStatsProcessor):
    documents = "s3://ai2-llm/pretraining-data/sources/common-crawl/v1/documents/cc_en_*/*.gz"
    stats = "s3://ai2-llm/stats/olmo-mix/v1/cc/v1/**/*.gz"

    @classmethod
    def process_single(
        cls, source_path: str, destination_path: str, queue: "Queue[Union[Tuple[int, ...], None]]", **kwargs: Any
    ):
        attributes = [source_path.replace("/documents/", "/attributes/c4_rules/")]

        # for the data sheet, what statistics you think we should include? I could
        # do # of docs, # tokens, distribution of URLs, pronouns, s2 FOS, stack languages?
        doc_decoder = msgspec.json.Decoder(InputSpec)
        attr_decoder = msgspec.json.Decoder(OutputSpec)
        stats = {
            "length": 0,
            "count": 0,
            "c4_count": 0,
            "c4_length": 0,
            "c4_matches": 0,
        }
        documents = 0
        interval = 10_000

        with ExitStack() as stack:
            doc_file = stack.enter_context(smart_open.open(source_path, "rb"))
            atts_files = [stack.enter_context(smart_open.open(path, "rb")) for path in attributes]

            for doc_line, *attr_lines in zip(doc_file, *atts_files):
                doc = doc_decoder.decode(doc_line)
                stats["length"] += len(doc.text)
                stats["count"] += 1

                attrs = {}
                for line in attr_lines:
                    attrs.update(attr_decoder.decode(line).attributes)

                # C4 stats
                c4_removal = attrs.get("c4_rules__c4_v1__lines_with_no_ending_punctuation", [])
                stats["c4_count"] += len(c4_removal)
                stats["c4_length"] += sum(s[-1] for s in c4_removal)
                stats["c4_matches"] += 1 if c4_removal else 0

                documents += 1

                if documents % interval == 0:
                    cls.increment_progressbar(queue, documents=interval)

        cls.increment_progressbar(queue, files=1, documents=documents % interval)

        with smart_open.open(destination_path, "wt") as destination_file:
            destination_file.write(json.dumps(stats, indent=2))


@Registry.add
class just_cc_dedup(BaseStatsProcessor):
    documents = "s3://ai2-llm/pretraining-data/sources/common-crawl/v1-c4-cleaned/documents/cc_en_*/*.gz"
    stats = "s3://ai2-llm/stats/olmo-mix/v1/cc/just_cc_dedup/**/*.gz"

    @classmethod
    def process_single(
        cls, source_path: str, destination_path: str, queue: "Queue[Union[Tuple[int, ...], None]]", **kwargs: Any
    ):
        dedup = source_path.replace("/documents/", "/attributes/dedupe_paragraphs/")

        attr_decoder = msgspec.json.Decoder(OutputSpec)
        stats = {
            "dedupe_paragraphs_count": 0,
            "dedupe_paragraphs_length": 0,
            "dedupe_paragraphs_matches": 0,
        }
        documents = 0
        interval = 10_000

        with ExitStack() as stack:
            try:
                dedup_file = stack.enter_context(smart_open.open(dedup, "rb"))
            except Exception:
                return

            for ln in dedup_file:
                attrs = attr_decoder.decode(ln).attributes

                # Duplicates stats
                dups = [p for p in attrs.get("bff_duplicate_paragraph_spans", []) if p[1] - p[0] > 0]
                stats["dedupe_paragraphs_count"] += len(dups)
                stats["dedupe_paragraphs_length"] += sum(s[1] - s[0] for s in dups)
                stats["dedupe_paragraphs_matches"] += 1 if dups else 0

                documents += 1

                if documents % interval == 0:
                    cls.increment_progressbar(queue, documents=interval)

        cls.increment_progressbar(queue, files=1, documents=documents % interval)

        with smart_open.open(destination_path, "wt") as destination_file:
            destination_file.write(json.dumps(stats, indent=2))


@Registry.add
class cc_v1_c4_cleaned(BaseStatsProcessor):
    documents = "s3://ai2-llm/pretraining-data/sources/common-crawl/v1-c4-cleaned/documents/cc_en_*/*.gz"
    stats = "s3://ai2-llm/stats/olmo-mix/v1/cc/v1_c4_cleaned/**/*.gz"
    decontamination_key: str = "decontamination"

    @classmethod
    def gopher_rules(cls, attrs: Dict[str, List[Tuple[int, int, float]]]) -> List[Tuple[int, int, float]]:
        matching_spans: List[Tuple[int, int, float]] = []

        for span in attrs.get("gopher_rules__gopher_v1__word_count", []):
            if span[2] < 50 or span[2] > 100000:
                bisect.insort(matching_spans, (span[0], span[1], 1.0))

        for span in attrs.get("gopher_rules__gopher_v1__median_word_length", []):
            if span[2] < 3 or span[2] > 10:
                bisect.insort(matching_spans, (span[0], span[1], 1.0))

        for span in attrs.get("gopher_rules__gopher_v1__symbol_to_word_ratio", []):
            if span[2] > 0.1:
                bisect.insort(matching_spans, (span[0], span[1], 1.0))

        for span in attrs.get("gopher_rules__gopher_v1__fraction_of_words_with_alpha_character", []):
            if span[2] < 0.8:
                bisect.insort(matching_spans, (span[0], span[1], 1.0))

        for span in attrs.get("gopher_rules__gopher_v1__required_word_count", []):
            if span[2] < 2:
                bisect.insort(matching_spans, (span[0], span[1], 1.0))

        for span in attrs.get("gopher_rules__gopher_v1__fraction_of_lines_starting_with_bullet_point", []):
            if span[2] > 0.9:
                bisect.insort(matching_spans, (span[0], span[1], 1.0))

        for span in attrs.get("gopher_rules__gopher_v1__fraction_of_lines_ending_with_ellipsis", []):
            if span[2] > 0.3:
                bisect.insort(matching_spans, (span[0], span[1], 1.0))

        for span in attrs.get("gopher_rules__gopher_v1__fraction_of_duplicate_lines", []):
            if span[2] > 0.3:
                bisect.insort(matching_spans, (span[0], span[1], 1.0))

        for span in attrs.get("gopher_rules__gopher_v1__fraction_of_characters_in_duplicate_lines", []):
            if span[2] > 0.3:
                bisect.insort(matching_spans, (span[0], span[1], 1.0))

        for span in attrs.get("gopher_rules__gopher_v1__fraction_of_characters_in_most_common_2gram", []):
            if span[2] > 0.2:
                bisect.insort(matching_spans, (span[0], span[1], 1.0))

        for span in attrs.get("gopher_rules__gopher_v1__fraction_of_characters_in_most_common_3gram", []):
            if span[2] > 0.18:
                bisect.insort(matching_spans, (span[0], span[1], 1.0))

        for span in attrs.get("gopher_rules__gopher_v1__fraction_of_characters_in_most_common_4gram", []):
            if span[2] > 0.16:
                bisect.insort(matching_spans, (span[0], span[1], 1.0))

        for span in attrs.get("gopher_rules__gopher_v1__fraction_of_characters_in_duplicate_5grams", []):
            if span[2] > 0.15:
                bisect.insort(matching_spans, (span[0], span[1], 1.0))

        for span in attrs.get("gopher_rules__gopher_v1__fraction_of_characters_in_duplicate_6grams", []):
            if span[2] > 0.14:
                bisect.insort(matching_spans, (span[0], span[1], 1.0))

        for span in attrs.get("gopher_rules__gopher_v1__fraction_of_characters_in_duplicate_7grams", []):
            if span[2] > 0.13:
                bisect.insort(matching_spans, (span[0], span[1], 1.0))

        for span in attrs.get("gopher_rules__gopher_v1__fraction_of_characters_in_duplicate_8grams", []):
            if span[2] > 0.12:
                bisect.insort(matching_spans, (span[0], span[1], 1.0))

        for span in attrs.get("gopher_rules__gopher_v1__fraction_of_characters_in_duplicate_9grams", []):
            if span[2] > 0.11:
                bisect.insort(matching_spans, (span[0], span[1], 1.0))

        for span in attrs.get("gopher_rules__gopher_v1__fraction_of_characters_in_duplicate_10grams", []):
            if span[2] > 0.10:
                bisect.insort(matching_spans, (span[0], span[1], 1.0))

        return cls._merge_spans(matching_spans)

    @classmethod
    def _merge_spans(cls, matching_spans: List[Tuple[int, int, float]]) -> List[Tuple[int, int, float]]:
        # merge spans if overlapping
        merged_spans: List[Tuple[int, int, float]] = []
        current_span = None

        for span in matching_spans:
            if span[1] - span[0] <= 0:
                continue
            elif current_span is None:
                current_span = span
            elif span[0] <= current_span[1]:  # type: ignore
                current_span = (current_span[0], span[1], 1.0)
            else:
                merged_spans.append(current_span)
                current_span = span

        if current_span:
            merged_spans.append(current_span)

        return merged_spans

    @classmethod
    def process_single(
        cls, source_path: str, destination_path: str, queue: "Queue[Union[Tuple[int, ...], None]]", **kwargs: Any
    ):
        attributes = [
            source_path.replace("/documents/", "/attributes/gopher_rules/"),
            source_path.replace("/documents/", f"/attributes/{cls.decontamination_key}/"),
            source_path.replace("/documents/", "/attributes/hatespeech_nsfw_cc_v3/"),
            source_path.replace("/documents/", "/attributes/pii_detection/"),
            source_path.replace("/documents/", "/attributes/dedupe_paragraphs/"),
        ]

        doc_decoder = msgspec.json.Decoder(InputSpec)
        attr_decoder = msgspec.json.Decoder(OutputSpec)
        stats = {
            "length": 0,
            "count": 0,
            "gopher_count": 0,
            "gopher_length": 0,
            "gopher_matches": 0,
            "decontamination_count": 0,
            "decontamination_length": 0,
            "decontamination_matches": 0,
            "dedupe_paragraphs_count": 0,
            "dedupe_paragraphs_length": 0,
            "dedupe_paragraphs_matches": 0,
            "hatespeech_nsfw_count": 0,
            "hatespeech_nsfw_length": 0,
            "hatespeech_nsfw_matches": 0,
            "pii_count": 0,
            "pii_length": 0,
            "pii_matches_le_5": 0,
            "pii_matches_gt_5": 0,
        }
        documents = 0
        interval = 10_000

        with ExitStack() as stack:
            doc_file = stack.enter_context(smart_open.open(source_path, "rb"))

            try:
                atts_files = [stack.enter_context(smart_open.open(path, "rb")) for path in attributes]
            except Exception:
                return

            for doc_line, *attr_lines in zip(doc_file, *atts_files):
                doc = doc_decoder.decode(doc_line)
                stats["length"] += len(doc.text)
                stats["count"] += 1

                attrs = {}
                for line in attr_lines:
                    attrs.update(attr_decoder.decode(line).attributes)

                # Gopher stats
                gopher_removal = cls.gopher_rules(attrs)
                stats["gopher_count"] += len(gopher_removal)
                stats["gopher_length"] += sum(s[1] - s[0] for s in gopher_removal)
                stats["gopher_matches"] += 1 if gopher_removal else 0

                # Deduplication stats
                decontamination_removal = attrs.get("bff_duplicate_paragraph_spans_decontamination", [])
                stats["decontamination_count"] += len(decontamination_removal)
                stats["decontamination_length"] += sum(s[1] - s[0] for s in decontamination_removal)
                stats["decontamination_matches"] += 1 if decontamination_removal else 0

                # jigsaw stats
                jigsaw_match: List[Tuple[int, int, float]] = []
                nsfw = attrs.get("hatespeech_nsfw_cc_v3__jigsaw_nsfw_sencence_v2____label__nsfw", [])
                for span in nsfw:
                    if span[2] > 0.4:
                        bisect.insort(jigsaw_match, (span[0], span[1], 1.0))

                toxic = attrs.get("hatespeech_nsfw_cc_v3__jigsaw_hatespeech_sentence_v2____label__toxic", [])
                for span in toxic:
                    if span[2] > 0.4:
                        bisect.insort(jigsaw_match, (span[0], span[1], 1.0))

                jigsaw_match = cls._merge_spans(jigsaw_match)

                stats["hatespeech_nsfw_count"] += len(jigsaw_match)
                stats["hatespeech_nsfw_length"] += sum(s[1] - s[0] for s in jigsaw_match)
                stats["hatespeech_nsfw_matches"] += 1 if jigsaw_match else 0

                # PII stats
                pii_removal = (
                    attrs.get("pii_detection__pii_regex_with_counts_fast_v2__EMAIL_ADDRESS", [])
                    + attrs.get("pii_detection__pii_regex_with_counts_fast_v2__PHONE_NUMBER", [])
                    + attrs.get("pii_detection__pii_regex_with_counts_fast_v2__IP_ADDRESS", [])
                )
                stats["pii_count"] += len(pii_removal)
                stats["pii_length"] += sum(s[1] - s[0] for s in pii_removal)
                stats["pii_matches_le_5"] += 1 if 0 < len(pii_removal) <= 5 else 0
                stats["pii_matches_gt_5"] += 1 if len(pii_removal) > 5 else 0

                # Duplicates stats
                dups = [p for p in attrs.get("bff_duplicate_paragraph_spans", []) if p[1] - p[0] > 0]
                stats["dedupe_paragraphs_count"] += len(dups)
                stats["dedupe_paragraphs_length"] += sum(s[1] - s[0] for s in dups)
                stats["dedupe_paragraphs_matches"] += 1 if dups else 0

                documents += 1

                if documents % interval == 0:
                    cls.increment_progressbar(queue, documents=interval)

        cls.increment_progressbar(queue, files=1, documents=documents % interval)

        with smart_open.open(destination_path, "wt") as destination_file:
            destination_file.write(json.dumps(stats, indent=2))


@Registry.add
class v15_cc_c4_cleaned(cc_v1_c4_cleaned):
    documents = "s3://ai2-llm/pretraining-data/sources/common-crawl/v1-c4-cleaned/documents/cc_en_*/*.gz"
    stats = "s3://ai2-llm/stats/olmo-mix/v15/cc/v1_c4_cleaned/**/*.gz"
    decontamination_key: str = "perplexity_suite_v3_option2"

@Registry.add
class LineStatsCC(cc_v1_c4_cleaned):
    # Selection of documents:
    # import random; print([random.randint(0, 1334) for _ in range(10)])
    documents = [
        "s3://ai2-llm/pretraining-data/sources/common-crawl/v1-c4-cleaned/documents/cc_en_head/cc_en_head-0700.json.gz",
        "s3://ai2-llm/pretraining-data/sources/common-crawl/v1-c4-cleaned/documents/cc_en_head/cc_en_head-0724.json.gz",
        "s3://ai2-llm/pretraining-data/sources/common-crawl/v1-c4-cleaned/documents/cc_en_head/cc_en_head-0788.json.gz",
        "s3://ai2-llm/pretraining-data/sources/common-crawl/v1-c4-cleaned/documents/cc_en_head/cc_en_head-1286.json.gz",
        "s3://ai2-llm/pretraining-data/sources/common-crawl/v1-c4-cleaned/documents/cc_en_head/cc_en_head-0600.json.gz",
        "s3://ai2-llm/pretraining-data/sources/common-crawl/v1-c4-cleaned/documents/cc_en_head/cc_en_head-0752.json.gz",
        "s3://ai2-llm/pretraining-data/sources/common-crawl/v1-c4-cleaned/documents/cc_en_head/cc_en_head-0239.json.gz",
        "s3://ai2-llm/pretraining-data/sources/common-crawl/v1-c4-cleaned/documents/cc_en_head/cc_en_head-1270.json.gz",
        "s3://ai2-llm/pretraining-data/sources/common-crawl/v1-c4-cleaned/documents/cc_en_head/cc_en_head-0786.json.gz",
        "s3://ai2-llm/pretraining-data/sources/common-crawl/v1-c4-cleaned/documents/cc_en_head/cc_en_head-0857.json.gz",
    ]
    stats = [
        "./cc_en_head-0700-stats.json.gz",
        "./cc_en_head-0724-stats.json.gz",
        "./cc_en_head-0788-stats.json.gz",
        "./cc_en_head-1286-stats.json.gz",
        "./cc_en_head-0600-stats.json.gz",
        "./cc_en_head-0752-stats.json.gz",
        "./cc_en_head-0239-stats.json.gz",
        "./cc_en_head-1270-stats.json.gz",
        "./cc_en_head-0786-stats.json.gz",
        "./cc_en_head-0857-stats.json.gz",
    ]
    decontamination_key: str = 'decontamination'

    @classmethod
    def cli(cls, num_workers: int = 1, debug: bool = False, **process_single_kwargs: Any) -> None:
        cls._run_parallel_processor(
            stats_root=cls.stats,
            num_workers=num_workers,
            debug=debug,
            **process_single_kwargs,
        )
            
    @classmethod
    def process_single(
        cls, source_path: str, destination_path: str, queue: "Queue[Union[Tuple[int, ...], None]]", **kwargs: Any
    ):
        attributes = [
            source_path.replace("/documents/", "/attributes/gopher_rules/"),
            source_path.replace("/documents/", f"/attributes/{cls.decontamination_key}/"),
            source_path.replace("/documents/", "/attributes/hatespeech_nsfw_cc_v3/"),
            source_path.replace("/documents/", "/attributes/pii_detection/"),
            source_path.replace("/documents/", "/attributes/dedupe_paragraphs/"),
        ]

        doc_decoder = msgspec.json.Decoder(InputSpec)
        attr_decoder = msgspec.json.Decoder(OutputSpec)
        documents = 0
        interval = 10_000

        with ExitStack() as stack:
            doc_file = stack.enter_context(smart_open.open(source_path, "rb"))
            out_file = stack.enter_context(smart_open.open(destination_path, "wt"))

            try:
                atts_files = [stack.enter_context(smart_open.open(path, "rb")) for path in attributes]
            except Exception:
                return

            for doc_line, *attr_lines in zip(doc_file, *atts_files):
                doc = doc_decoder.decode(doc_line)
                attrs = {}
                for line in attr_lines:
                    attrs.update(attr_decoder.decode(line).attributes)
                out_line = {}

                # Gopher stats
                gopher_removal = cls.gopher_rules(attrs)
                out_line["gopher_spans"] = gopher_removal

                # Deduplication stats
                decontamination_removal = attrs.get("bff_duplicate_paragraph_spans_decontamination", [])
                out_line["decontamination_spans"] = decontamination_removal

                # jigsaw stats
                jigsaw_match: List[Tuple[int, int, float]] = []
                nsfw = attrs.get("hatespeech_nsfw_cc_v3__jigsaw_nsfw_sencence_v2____label__nsfw", [])
                for span in nsfw:
                    if span[2] > 0.4:
                        bisect.insort(jigsaw_match, (span[0], span[1], 1.0))

                toxic = attrs.get("hatespeech_nsfw_cc_v3__jigsaw_hatespeech_sentence_v2____label__toxic", [])
                for span in toxic:
                    if span[2] > 0.4:
                        bisect.insort(jigsaw_match, (span[0], span[1], 1.0))

                jigsaw_match = cls._merge_spans(jigsaw_match)
                out_line["hatespeech_spans"] = jigsaw_match

                # PII stats
                pii_removal = (
                    attrs.get("pii_detection__pii_regex_with_counts_fast_v2__EMAIL_ADDRESS", [])
                    + attrs.get("pii_detection__pii_regex_with_counts_fast_v2__PHONE_NUMBER", [])
                    + attrs.get("pii_detection__pii_regex_with_counts_fast_v2__IP_ADDRESS", [])
                )
                out_line["pii_spans"] = pii_removal

                # Duplicates stats
                dups = [p for p in attrs.get("bff_duplicate_paragraph_spans", []) if p[1] - p[0] > 0]
                out_line["dedupe_paragraphs_spans"] = dups

                documents += 1

                if documents % interval == 0:
                    cls.increment_progressbar(queue, documents=interval)

                out_file.write(json.dumps(out_line) + "\n")

        cls.increment_progressbar(queue, files=1, documents=documents % interval)

class C4InputSpec(InputSpec):
    metadata: Dict[str, Any] = msgspec.field(default_factory=dict)


@Registry.add
class c4(BaseStatsProcessor):
    @classmethod
    def process_single(
        cls, source_path: str, destination_path: str, queue: "Queue[Union[Tuple[int, ...], None]]", **kwargs: Any
    ):
        attrs_path = source_path.replace("/documents/", "/attributes/decontamination/")

        documents_decoder = msgspec.json.Decoder(C4InputSpec)
        attributes_decoder = msgspec.json.Decoder(OutputSpec)
        counts = Counts()
        interval = 10_000

        with smart_open.open(source_path, "rb") as doc_file, smart_open.open(attrs_path, "rb") as attrs_file:
            for source_line, attributes_line in zip(doc_file, attrs_file):
                document = documents_decoder.decode(source_line)
                attributes = attributes_decoder.decode(attributes_line)

                text = document.text
                for start, end, _ in sorted(
                    attributes.attributes.get("bff_duplicate_paragraph_spans_decontamination", []),
                    key=lambda t: -t[1],
                ):
                    # remove duplicate
                    text = text[:start] + text[end:]

                counts.add(text=text, url=document.metadata["url"])

                if counts.documents % interval == 0:
                    cls.increment_progressbar(queue, documents=interval)

        cls.increment_progressbar(queue, files=1, documents=counts.documents % interval)

        with smart_open.open(destination_path, "wt") as destination_file:
            destination_file.write(json.dumps(counts.to_dict(), indent=2))

    @classmethod
    def cli(cls, num_workers: int = 1, debug: bool = False, **process_single_kwargs: Any) -> None:
        with TemporaryDirectory() as tempdir:
            documents = "s3://ai2-llm/pretraining-data/sources/c4/v0/documents/train/*.gz"
            stats = "s3://ai2-llm/stats/olmo-mix/v1/web/c4"
            metadata = os.path.join(tempdir, "c4")

            processor = cls(
                source_prefix=documents,
                destination_prefix=stats,
                metadata_prefix=metadata,
                num_processes=num_workers,
                debug=debug,
            )
            processor(**process_single_kwargs)


@Registry.add
class s2(BaseStatsProcessor):
    @classmethod
    def process_single(
        cls, source_path: str, destination_path: str, queue: "Queue[Union[Tuple[int, ...], None]]", **kwargs: Any
    ):
        attrs_path = source_path.replace("/documents/", "/attributes/decontamination/")

        documents_decoder = msgspec.json.Decoder(C4InputSpec)
        attributes_decoder = msgspec.json.Decoder(OutputSpec)

        interval = 10_000

        counts: dict = {
            f: {"year": {}, "s2fos": {}, "documents": 0, "tokens": 0} for f in ["full_text", "abstract"]
        }
        key = "full_text" if "dataset=s2orc" in source_path else "abstract"
        cnt = 0

        with smart_open.open(source_path, "rb") as doc_file, smart_open.open(attrs_path, "rb") as attrs_file:
            for source_line, attributes_line in zip(doc_file, attrs_file):
                cnt += 1

                document = documents_decoder.decode(source_line)
                attributes = attributes_decoder.decode(attributes_line)

                text = document.text
                for start, end, _ in sorted(
                    attributes.attributes.get("bff_duplicate_paragraph_spans_decontamination", []),
                    key=lambda t: -t[1],
                ):
                    # remove duplicate
                    text = text[:start] + text[end:]

                if not (text := text.strip()):
                    continue

                counts[key]["documents"] += 1
                counts[key]["tokens"] += len(blingfire.text_to_words(text).split())

                if document.metadata["year"] not in counts[key]["year"]:
                    counts[key]["year"][document.metadata["year"]] = 0
                counts[key]["year"][document.metadata["year"]] += 1

                if len(document.metadata["s2fieldsofstudy"]) == 0:
                    document.metadata["s2fieldsofstudy"] = ["null"]

                for fos in document.metadata["s2fieldsofstudy"]:
                    if fos not in counts[key]["s2fos"]:
                        counts[key]["s2fos"][fos] = 0
                    counts[key]["s2fos"][fos] += 1

                if cnt % interval == 0:
                    cls.increment_progressbar(queue, documents=interval)

        cls.increment_progressbar(queue, files=1, documents=cnt % interval)

        with smart_open.open(destination_path, "wt") as destination_file:
            destination_file.write(json.dumps(counts, indent=2))

    @classmethod
    def cli(cls, num_workers: int = 1, debug: bool = False, **process_single_kwargs: Any) -> None:
        with TemporaryDirectory() as tempdir:
            documents = (
                "s3://ai2-llm/pretraining-data/sources/s2/v3-fos/documents/dataset=*/split=train/part_id=*/*.gz"
            )
            stats = "s3://ai2-llm/stats/olmo-mix/v1/papers/peS2o"
            metadata = os.path.join(tempdir, "s2")

            processor = cls(
                source_prefix=documents,
                destination_prefix=stats,
                metadata_prefix=metadata,
                num_processes=num_workers,
                debug=debug,
            )
            processor(**process_single_kwargs)


class StackInputSpec(InputSpec):
    metadata: Dict[str, Any] = msgspec.field(default_factory=dict)
    attributes: Dict[str, Any] = msgspec.field(default_factory=dict)


@Registry.add
class stack_v2(BaseStatsProcessor):
    documents = "s3://ai2-llm/pretraining-data/sources/stack-dedup/v1-mixer/documents/*.gz"
    stats = "s3://ai2-llm/stats/olmo-mix/v1/stack/v1-mixer/*.gz"

    @classmethod
    def process_single(
        cls, source_path: str, destination_path: str, queue: "Queue[Union[Tuple[int, ...], None]]", **kwargs: Any
    ):
        attrs_basic = source_path.replace("/documents/", "/attributes/basic/")
        attrs_code_secrets = source_path.replace("/documents/", "/attributes/rpj-heuristics/")
        # attrs_dedupe_documents = source_path.replace("/documents/", "/attributes/dedupe_documents/")
        # attrs_pii = source_path.replace("/documents/", "/attributes/pii/")

        documents_decoder = msgspec.json.Decoder(StackInputSpec)
        attributes_decoder = msgspec.json.Decoder(OutputSpec)

        interval = 10_000

        counts: dict = {
            "extension": defaultdict(int),
            "license": defaultdict(int),
            "length": 0,
            "count": 0,
            "rpj_count": 0,
            "rpj_length": 0,
            "rpj_matches": 0,
        }
        cnt = 0

        with ExitStack() as stack:
            doc_file = stack.enter_context(smart_open.open(source_path, "rb"))
            attributes_files = [
                stack.enter_context(smart_open.open(attrs_basic, "rb")),
                stack.enter_context(smart_open.open(attrs_code_secrets, "rb")),
                # stack.enter_context(smart_open.open(attrs_dedupe_documents, "rb")),
                # stack.enter_context(smart_open.open(attrs_pii, "rb")),
            ]

            # with smart_open.open(source_path, "rb") as doc_file, smart_open.open(attrs_path, "rb") as attrs_file:
            for source_line, *attributes_line in zip(doc_file, *attributes_files):
                cnt += 1

                doc = documents_decoder.decode(source_line)
                for ln in attributes_line:
                    attributes = attributes_decoder.decode(ln)
                    doc.attributes.update(attributes.attributes)

                if doc.attributes["basic__random_number_v1__random"][-1][-1] > 0.996:
                    # test set; see
                    # https://github.com/allenai/LLM/blob/642d0fad3fb2efd816af507250c4c65c8678cb44/pretrain_data/the_stack/v2-mixer/ablations/v2-mixer-held-out.json#L15
                    continue

                spans = []
                if doc.attributes["rpj_heuristics__code_redpajama_taggers_v1__max_line_length_doc"][0][2] > 1000:
                    spans.append((0, len(doc.text), 1.0))
                elif doc.attributes["rpj_heuristics__code_redpajama_taggers_v1__avg_line_length_doc"][0][2] > 100:
                    spans.append((0, len(doc.text), 1.0))
                elif doc.attributes["rpj_heuristics__code_redpajama_taggers_v1__alnum_prop_doc"][0][2] < 0.25:
                    spans.append((0, len(doc.text), 1.0))
                elif doc.attributes["rpj_heuristics__code_redpajama_taggers_v1__alpha_token_prop_doc"][0][2] < 1.5:
                    spans.append((0, len(doc.text), 1.0))

                counts["rpj_count"] += len(spans)
                counts["rpj_length"] += sum(s[1] - s[0] for s in spans)
                counts["rpj_matches"] += 1 if spans else 0

                counts["count"] += 1
                counts["length"] += len(doc.text)

                counts["extension"][doc.metadata["ext"]] += 1
                for license in doc.metadata["max_forks_repo_licenses"]:
                    counts["license"][license] += 1

                cnt += 1

                if cnt % interval == 0:
                    cls.increment_progressbar(queue, documents=interval)

        cls.increment_progressbar(queue, files=1, documents=cnt % interval)

        with smart_open.open(destination_path, "wt") as destination_file:
            destination_file.write(json.dumps(counts, indent=2))


@Registry.add
class stack_v3(stack_v2):
    documents = "s3://ai2-llm/pretraining-data/sources/stack-dedup/v2-mixer/documents/*.gz"
    stats = "s3://ai2-llm/stats/olmo-mix/v1/stack/v2-mixer/*.gz"

    @classmethod
    def process_single(
        cls, source_path: str, destination_path: str, queue: "Queue[Union[Tuple[int, ...], None]]", **kwargs: Any
    ):
        attrs_basic = source_path.replace("/documents/", "/attributes/basic/")
        # attrs_code_secrets = source_path.replace("/documents/", "/attributes/rpj-heuristics/")
        attrs_dedupe_documents = source_path.replace("/documents/", "/attributes/dedupe_documents/")
        attrs_pii = source_path.replace("/documents/", "/attributes/pii/")

        documents_decoder = msgspec.json.Decoder(StackInputSpec)
        attributes_decoder = msgspec.json.Decoder(OutputSpec)

        interval = 10_000

        counts: dict = {
            "extension": defaultdict(int),
            "license": defaultdict(int),
            "length": 0,
            "count": 0,
            "rpj_count": 0,
            "rpj_length": 0,
            "rpj_matches": 0,
        }
        cnt = 0

        with ExitStack() as stack:
            doc_file = stack.enter_context(smart_open.open(source_path, "rb"))
            attributes_files = [
                stack.enter_context(smart_open.open(attrs_basic, "rb")),
                stack.enter_context(smart_open.open(attrs_dedupe_documents, "rb")),
                stack.enter_context(smart_open.open(attrs_pii, "rb")),
            ]

            # with smart_open.open(source_path, "rb") as doc_file, smart_open.open(attrs_path, "rb") as attrs_file:
            for source_line, *attributes_line in zip(doc_file, *attributes_files):
                cnt += 1

                doc = documents_decoder.decode(source_line)
                for ln in attributes_line:
                    attributes = attributes_decoder.decode(ln)
                    doc.attributes.update(attributes.attributes)

                if doc.attributes["basic__random_number_v1__random"][-1][-1] > 0.996:
                    # test set; see
                    # https://github.com/allenai/LLM/blob/642d0fad3fb2efd816af507250c4c65c8678cb44/pretrain_data/the_stack/v2-mixer/ablations/v2-mixer-held-out.json#L15
                    continue

                spans = []
                if doc.attributes["rpj_heuristics__code_redpajama_taggers_v1__max_line_length_doc"][0][2] > 1000:
                    spans.append((0, len(doc.text), 1.0))
                elif doc.attributes["rpj_heuristics__code_redpajama_taggers_v1__avg_line_length_doc"][0][2] > 100:
                    spans.append((0, len(doc.text), 1.0))
                elif doc.attributes["rpj_heuristics__code_redpajama_taggers_v1__alnum_prop_doc"][0][2] < 0.25:
                    spans.append((0, len(doc.text), 1.0))
                elif doc.attributes["rpj_heuristics__code_redpajama_taggers_v1__alpha_token_prop_doc"][0][2] < 1.5:
                    spans.append((0, len(doc.text), 1.0))

                counts["rpj_count"] += len(spans)
                counts["rpj_length"] += sum(s[1] - s[0] for s in spans)
                counts["rpj_matches"] += 1 if spans else 0

                counts["count"] += 1
                counts["length"] += len(doc.text)

                counts["extension"][doc.metadata["ext"]] += 1
                for license in doc.metadata["max_forks_repo_licenses"]:
                    counts["license"][license] += 1

                cnt += 1

                if cnt % interval == 0:
                    cls.increment_progressbar(queue, documents=interval)

        cls.increment_progressbar(queue, files=1, documents=cnt % interval)

        with smart_open.open(destination_path, "wt") as destination_file:
            destination_file.write(json.dumps(counts, indent=2))


@Registry.add
class stack_v4(stack_v2):
    documents = "s3://ai2-llm/pretraining-data/sources/stack-dedup/v4-train/documents/*/*.gz"
    stats = "s3://ai2-llm/stats/olmo-mix/v1_5/stack/v4-train/**/*.gz"

    @classmethod
    def process_single(
        cls, source_path: str, destination_path: str, queue: "Queue[Union[Tuple[int, ...], None]]", **kwargs: Any
    ):
        attrs_basic = source_path.replace("/documents/", "/attributes/perplexity_suite_v3_option2/")
        # attrs_code_secrets = source_path.replace("/documents/", "/attributes/rpj-heuristics/")
        # attrs_dedupe_documents = source_path.replace("/documents/", "/attributes/dedupe_documents/")
        # attrs_pii = source_path.replace("/documents/", "/attributes/pii/")

        documents_decoder = msgspec.json.Decoder(StackInputSpec)
        attributes_decoder = msgspec.json.Decoder(OutputSpec)

        interval = 10_000

        counts: dict = {
            "extension": defaultdict(int),
            "license": defaultdict(int),
            "length": 0,
            "count": 0,
            "decontamination_count": 0,
            "decontamination_length": 0,
            "decontamination_matches": 0,
        }
        cnt = 0

        with ExitStack() as stack:
            doc_file = stack.enter_context(smart_open.open(source_path, "rb"))
            attributes_files = [
                stack.enter_context(smart_open.open(attrs_basic, "rb")),
            ]

            # with smart_open.open(source_path, "rb") as doc_file, smart_open.open(attrs_path, "rb") as attrs_file:
            for source_line, *attributes_line in zip(doc_file, *attributes_files):
                cnt += 1

                doc = documents_decoder.decode(source_line)
                for ln in attributes_line:
                    attributes = attributes_decoder.decode(ln)
                    doc.attributes.update(attributes.attributes)

                attrs = {}
                for line in attributes_line:
                    attrs.update(attributes_decoder.decode(line).attributes)

                # Deduplication stats
                decontamination_removal = attrs.get("bff_duplicate_paragraph_spans_decontamination", [])
                counts["decontamination_count"] += len(decontamination_removal)
                counts["decontamination_length"] += sum(s[1] - s[0] for s in decontamination_removal)
                counts["decontamination_matches"] += 1 if decontamination_removal else 0

                counts["count"] += 1
                counts["length"] += len(doc.text)

                counts["extension"][doc.metadata["ext"]] += 1
                for license in doc.metadata["max_forks_repo_licenses"]:
                    counts["license"][license] += 1

                cnt += 1

                if cnt % interval == 0:
                    cls.increment_progressbar(queue, documents=interval)

        cls.increment_progressbar(queue, files=1, documents=cnt % interval)

        with smart_open.open(destination_path, "wt") as destination_file:
            destination_file.write(json.dumps(counts, indent=2))


@Registry.add
class LineStatsStack(BaseStatsProcessor):
    # Testing
    #documents = "s3://ai2-llm/pretraining-data/sources/stack-dedup/v0/documents/abap/data_0000.jsonl.gz"
    #stats = "/data/niklas/dolma/abap/data_0000.json.gz"
    documents = "s3://ai2-llm/pretraining-data/sources/stack-dedup/v0/documents/*/*.gz"
    stats = "/data/niklas/dolma/stack"

    @classmethod
    def cli(cls, num_workers: int = 1, debug: bool = False, **process_single_kwargs: Any) -> None:
        stats_root = cls.stats

        cls._run_parallel_processor(
            stats_root=stats_root,
            num_workers=num_workers,
            debug=debug,
            **process_single_kwargs,
        )

    @classmethod
    def process_single(
        cls, source_path: str, destination_path: str, queue: "Queue[Union[Tuple[int, ...], None]]", **kwargs: Any
    ):
        # E.g. `s3://ai2-llm/pretraining-data/sources/stack-dedup/v0/attributes/paper_analysis/abap/`
        attr_path = source_path.replace("/documents/", "/attributes/paper_analysis/")

        doc_decoder = msgspec.json.Decoder(InputSpec)
        attr_decoder = msgspec.json.Decoder(OutputSpec)
        documents = 0
        interval = 10_000

        with ExitStack() as stack:
            doc_file = stack.enter_context(smart_open.open(source_path, "rb"))
            out_file = stack.enter_context(smart_open.open(destination_path, "wt"))

            try:
                attr_file = stack.enter_context(smart_open.open(attr_path, "rb"))
            except Exception as e:
                print(e)
                return
            
            for doc_line, attrs in zip(doc_file, attr_file):
                doc = doc_decoder.decode(doc_line)
                attrs = attr_decoder.decode(attrs).attributes
                out_line = {}

                ## RPJ ##
                if (attrs["paper_analysis__code_redpajama_taggers_v1__max_line_length_doc"][0][2] > 1000) or \
                    (attrs["paper_analysis__code_redpajama_taggers_v1__avg_line_length_doc"][0][2] > 100) or \
                    (attrs["paper_analysis__code_redpajama_taggers_v1__alnum_prop_doc"][0][2] < 0.25) or \
                    (attrs["paper_analysis__code_redpajama_taggers_v1__alpha_token_prop_doc"][0][2] < 1.5):
                    out_line["rpj"] = 1
                else:
                    out_line["rpj"] = 0

                ## StarCoder ##
                if (attrs["paper_analysis__code_starcoder_taggers_v2__has_xml_template_doc"][0][2] > 0.0) or \
                (attrs["paper_analysis__code_starcoder_taggers_v2__code_to_comment_ratio_doc"][0][2] > 0.8) or \
                (attrs["paper_analysis__code_starcoder_taggers_v2__code_to_comment_ratio_doc"][0][2] <= 0.01) or \
                (
                    any(x in source_path for x in ["python", "java", "javascript"]) and \
                    (attrs["paper_analysis__code_starcoder_taggers_v2__code_to_text_ratio_html_doc"][0][2] <= 0.1)
                ):
                    out_line["starcoder"] = 1
                else:
                    out_line["starcoder"] = 0

                documents += 1

                if documents % interval == 0:
                    cls.increment_progressbar(queue, documents=interval)

                out_file.write(json.dumps(out_line) + "\n")

        cls.increment_progressbar(queue, files=1, documents=documents % interval)


@Registry.add
class decon_ppl_v3(BaseStatsProcessor):
    documents = [
        "s3://ai2-llm/pretraining-data/sources/gutenberg/v0/documents/*.gz",
        "s3://ai2-llm/pretraining-data/sources/c4/v0/documents/train/*.gz",
        "s3://ai2-llm/pretraining-data/sources/common-crawl/v1-c4-cleaned/documents/cc_en_head/*.gz",
        "s3://ai2-llm/pretraining-data/sources/common-crawl/v1-c4-cleaned/documents/cc_en_middle/*.gz",
        "s3://ai2-llm/pretraining-data/sources/common-crawl/v1-c4-cleaned/documents/cc_en_tail/*.gz",
        "s3://ai2-llm/pretraining-data/sources/s2/v3/documents/dataset=s2ag/split=train/*/*.gz",
        "s3://ai2-llm/pretraining-data/sources/s2/v3/documents/dataset=s2orc/split=train/*/*.gz",
        "s3://ai2-llm/pretraining-data/sources/reddit/v5-dedupe-pii-nsfw-toxic/documents/*.gz",
        "s3://ai2-llm/pretraining-data/sources/stack-dedup/v4-train/documents/*/*.gz",
        "s3://ai2-llm/pretraining-data/sources/wikipedia/v0/documents/lang=en/*.gz",
        "s3://ai2-llm/pretraining-data/sources/wikipedia/v0/documents/lang=simple/*.gz",
        "s3://ai2-llm/pretraining-data/sources/wikibooks/v0/documents/lang=en/*.gz",
        "s3://ai2-llm/pretraining-data/sources/wikibooks/v0/documents/lang=simple/*.gz",
    ]
    stats = "s3://ai2-llm/stats/olmo-mix/v1_5/decontamination_ppl_v3_option2"


# # # BELOW HERE: AGGREGATION # # #


@Registry.add
class web(BaseStatsProcessor):
    @staticmethod
    # read all paths in using threads
    def _read_json(path: str) -> Counts:
        with smart_open.open(path, "rt") as source_file:
            content = msgspec.json.decode(source_file.read())
            return Counts.from_dict(content)

    @classmethod
    def cli(cls, num_workers: int = 1, debug: bool = False, **process_single_kwargs: Any) -> None:
        paths = list(
            chain(
                glob_path("s3://ai2-llm/stats/olmo-mix/v1/web/c4/*"),
                glob_path("s3://ai2-llm/stats/olmo-mix/v1/web/common-crawl/**/*"),
            )
        )
        assert len(paths), "Run c4 and common-crawl first"

        with multiprocessing.Pool(num_workers) as pool:
            data = (cls._read_json(path) for path in paths) if debug else pool.imap(cls._read_json, paths)
            counts = Counts()

            for content in tqdm.tqdm(data, desc="Merging web stats", unit=" files", total=len(paths)):
                counts.merge(content, shrink=True)

        with smart_open.open("s3://ai2-llm/stats/olmo-mix/v1/web/summary.json", "wt") as destination_file:
            destination_file.write(json.dumps(counts.to_dict(), indent=2))


@Registry.add
class papers(BaseStatsProcessor):
    @staticmethod
    # read all paths in using threads
    def _read_json(path: str) -> dict:
        with smart_open.open(path, "rt") as source_file:
            return json.loads(source_file.read())

    @classmethod
    def cli(cls, num_workers: int = 1, debug: bool = False, **process_single_kwargs: Any) -> None:
        paths = list(glob_path("s3://ai2-llm/stats/olmo-mix/v1/papers/peS2o/**/*.gz"))
        assert len(paths), "Run s2 first"

        with multiprocessing.Pool(num_workers) as pool:
            data = (cls._read_json(path) for path in paths) if debug else pool.imap(cls._read_json, paths)
            counts: dict = {
                f: {"year": {}, "s2fos": {}, "documents": 0, "tokens": 0} for f in ["full_text", "abstract"]
            }
            for content in tqdm.tqdm(data, desc="Merging web stats", unit=" files", total=len(paths)):
                for key, values in content.items():
                    counts[key]["documents"] += values["documents"]
                    counts[key]["tokens"] += values["tokens"]

                    for year, count in values["year"].items():
                        if year not in counts[key]["year"]:
                            counts[key]["year"][year] = 0
                        counts[key]["year"][year] += count

                    for fos, count in values["s2fos"].items():
                        if fos not in counts[key]["s2fos"]:
                            counts[key]["s2fos"][fos] = 0
                        counts[key]["s2fos"][fos] += count

        with smart_open.open("s3://ai2-llm/stats/olmo-mix/v1/papers/summary.json", "wt") as destination_file:
            destination_file.write(json.dumps(counts, indent=2))


if __name__ == "__main__":
    os.environ["TOKENIZERS_PARALLELISM"] = "false"
    os.environ["PYTHONBREAKPOINT"] = "ipdb.set_trace"

    ap = argparse.ArgumentParser()
    ap.add_argument("stat", choices=[name for name, _ in Registry.all()])
    ap.add_argument("--debug", action="store_true")
    ap.add_argument("--num-workers", type=int, default=multiprocessing.cpu_count())
    args = ap.parse_args()

    Registry.get(args.stat).cli(num_workers=args.num_workers, debug=args.debug)



from unittest import TestCase

import numpy as np

from dolma.taggers.repetitions.utils import (
    find_end_first_consecutive_true,
    find_periodic_sequences,
    find_start_last_consecutive_true,
    group_consecutive_values,
)


class TestTrueLocsDetection(TestCase):
    def test_find_end_first_consecutive_true(self):
        arr = np.array([True, True, False, True])
        self.assertEqual(find_end_first_consecutive_true(arr), 2)

        arr = np.array([False, True])
        self.assertEqual(find_end_first_consecutive_true(arr), 0)

        arr = np.array([True, False])
        self.assertEqual(find_end_first_consecutive_true(arr), 1)

        arr = np.array([True])
        self.assertEqual(find_end_first_consecutive_true(arr), 1)

        arr = np.array([False])
        self.assertEqual(find_end_first_consecutive_true(arr), 0)

    def test_find_start_last_consecutive_true(self):
        arr = np.array([False, False, True, True])
        self.assertEqual(find_start_last_consecutive_true(arr), 2)

        arr = np.array([True, False, False, True])
        self.assertEqual(find_start_last_consecutive_true(arr), 3)

        arr = np.array([True, True, True, True])
        self.assertEqual(find_start_last_consecutive_true(arr), 0)

        arr = np.array([False, False, False, False])
        self.assertEqual(find_start_last_consecutive_true(arr), -1)

        arr = np.array([True, True, True, False])
        self.assertEqual(find_start_last_consecutive_true(arr), -1)

    def test_group_consecutive_values(self):
        arr = np.array([1, 2, 3, 4, 5])
        grouped = group_consecutive_values(arr)
        self.assertEqual(len(grouped), 1)
        self.assertTrue((grouped[0] == arr).all())

        arr = np.array([1, 2, 3, 5, 6, 7, 9, 10, 11])
        grouped = group_consecutive_values(arr)
        self.assertEqual(len(grouped), 3)
        self.assertTrue((grouped[0] == np.array([1, 2, 3])).all())
        self.assertTrue((grouped[1] == np.array([5, 6, 7])).all())
        self.assertTrue((grouped[2] == np.array([9, 10, 11])).all())

        arr = np.array([1, 3, 5, 7, 9])
        grouped = group_consecutive_values(arr)
        self.assertEqual(len(grouped), 5)
        for i in range(5):
            self.assertEqual(len(grouped[i]), 1)
            self.assertEqual(grouped[i][0], arr[i])

    def _to_array(self, s: str) -> np.ndarray:
        return np.array(list(map(int, s)))

    def test_find_periodic_sequences(self):
        # should be 0 repeated 5 times
        arr = self._to_array("5000007")
        sequences = list(find_periodic_sequences(arr, max_period=1))
        self.assertEqual(len(sequences), 1)
        self.assertEqual(sequences[0].start, 1)
        self.assertEqual(sequences[0].end, 6)
        self.assertEqual(sequences[0].period, 1)
        self.assertEqual(sequences[0].times, 5)

        # should be 01 repeated 3 times
        arr = self._to_array("501010127")
        sequences = list(find_periodic_sequences(arr, min_period=2, max_period=2))
        self.assertEqual(len(sequences), 1)
        self.assertEqual(sequences[0].start, 1)
        self.assertEqual(sequences[0].end, 7)
        self.assertEqual(sequences[0].period, 2)
        self.assertEqual(sequences[0].times, 3)

        # should be 0104 repeated 3 times
        arr = self._to_array("5301040104010401")
        sequences = list(find_periodic_sequences(arr, min_period=4, max_period=4))
        self.assertEqual(len(sequences), 1)
        self.assertEqual(sequences[0].start, 2)
        self.assertEqual(sequences[0].end, 16)
        self.assertEqual(sequences[0].period, 4)
        self.assertEqual(sequences[0].times, 3)

        # should be 040 repeated 4 times
        arr = np.array(list(map(int, "04004004004030")))
        sequences = list(find_periodic_sequences(arr, min_period=3, max_period=3))
        self.assertEqual(len(sequences), 1)
        self.assertEqual(sequences[0].start, 0)
        self.assertEqual(sequences[0].end, 12)
        self.assertEqual(sequences[0].period, 3)
        self.assertEqual(sequences[0].times, 4)

        # should have two repetitions: 46 repeated 4 times, and 550 repeated 3 times
        arr = np.array(list(map(int, "004646464639955055055046550")))
        sequences = list(find_periodic_sequences(arr, min_period=2, max_period=3))
        self.assertEqual(len(sequences), 2)
        self.assertEqual(sequences[0].start, 2)
        self.assertEqual(sequences[0].end, 10)
        self.assertEqual(sequences[0].period, 2)
        self.assertEqual(sequences[0].times, 4)
        self.assertEqual(sequences[1].start, 13)
        self.assertEqual(sequences[1].end, 22)
        self.assertEqual(sequences[1].period, 3)
        self.assertEqual(sequences[1].times, 3)

    def test_find_no_periodic_sequences(self):
        arr = np.array(list(map(int, "123456789")))
        sequences = list(find_periodic_sequences(arr, max_period=10))
        self.assertEqual(len(sequences), 0)

        arr = np.array(list(map(int, "112233445566778899")))
        sequences = list(find_periodic_sequences(arr, max_period=10))
        self.assertEqual(len(sequences), 0)



"""
Taggers to detect repetitions in the text.

@soldni
"""

import re
from abc import abstractmethod
from typing import Generator, List

import numpy as np
from tokenizers import Tokenizer

from ...core.data_types import DocResult, Document, Span
from ...core.registry import TaggerRegistry
from ...core.taggers import BaseTagger
from ...core.utils import split_paragraphs
from .utils import find_periodic_sequences


class BaseRepetitionsTagger(BaseTagger):
    @abstractmethod
    def _extract_from_text(self, text: str) -> Generator[Span, None, None]:
        raise NotImplementedError()

    def _extract_from_doc(self, doc: Document) -> Generator[Span, None, None]:
        yield from self._extract_from_text(doc.text)

    def _compute_document_stats(self, spans: List[Span], doc: Document) -> List[Span]:
        doc_max_span = Span(
            start=0,
            end=len(doc.text),
            type="doc_max_repetition",
            score=max(spans, key=lambda s: s.score).score if spans else 0.0,
        )
        doc_mean_reps_span = Span(
            start=0,
            end=len(doc.text),
            type="doc_mean_repetition",
            score=float(np.mean([s.score for s in spans]) if spans else 0),
        )
        doc_frac_reps_span = Span(
            start=0,
            end=len(doc.text),
            type="doc_frac_repetition",
            score=float(sum([s.score for s in spans]) / len(doc.text) if spans else 0),
        )
        return [doc_max_span, doc_mean_reps_span, doc_frac_reps_span]

    def predict(self, doc: Document) -> DocResult:
        """Predict method for the tagger."""
        reps_spans = list(self._extract_from_doc(doc))
        document_stats_spans = self._compute_document_stats(spans=reps_spans, doc=doc)
        return DocResult(doc=doc, spans=reps_spans + document_stats_spans)


@TaggerRegistry.add("repetitions_v1")
class RepetitionsTagger(BaseRepetitionsTagger):
    """Tagger to detect repetitions of of groups of characters.
    Only repetitions that occur at least 4 times are detected."""

    def __init__(self) -> None:
        self.re_char_repetitions = re.compile(r"(.+?)(\s?\1){3,}")
        super().__init__()

    def _extract_from_text(self, text: str) -> Generator[Span, None, None]:
        """Extract repetitions of characters in the text."""
        for match in self.re_char_repetitions.finditer(text):
            yield Span(
                start=(start := match.start()),
                end=(end := match.end()),
                type="repetition",
                score=float(end - start),
            )


@TaggerRegistry.add("paragraph_repetitions_v1")
class ParagraphRepetitionsTagger(RepetitionsTagger):
    """Tagger to detect repetitions of paragraphs.
    It's faster than the char repetition tagger, but it does not account for
    repetitions of characters that span multiple paragraphs."""

    def _extract_from_doc(self, doc: Document) -> Generator[Span, None, None]:
        offset = 0
        for paragraph in split_paragraphs(doc.text, remove_empty=False):
            for span in self._extract_from_text(paragraph.text):
                span.start += offset
                span.end += offset
                yield span
            offset += len(paragraph.text)


@TaggerRegistry.add("tokenizer_repetitions_v1")
class TokenizerRepetitionsTagger(BaseRepetitionsTagger):
    """Tagger to detect repetitions of tokens.

    It uses a tokenizer to split the text into tokens, and then identifies
    sequences of tokens that repeat at least 3 times."""

    TOKENIZER_IDENTIFIER = "allenai/eleuther-ai-gpt-neox-20b-pii-special"
    MIN_PERIOD = 1
    MAX_PERIOD = 13

    def __init__(self) -> None:
        self.tokenizer = Tokenizer.from_pretrained(self.TOKENIZER_IDENTIFIER)

    def _extract_from_text(self, text: str) -> Generator[Span, None, None]:
        tokens = self.tokenizer.encode(text, add_special_tokens=False)
        sequences_iter = find_periodic_sequences(
            arr=np.array(tokens.ids), min_period=self.MIN_PERIOD, max_period=self.MAX_PERIOD
        )
        for seq in sequences_iter:
            yield Span(
                start=(s := tokens.offsets[seq.start][0]),
                end=(e := tokens.offsets[seq.end - 1][1]),
                type="repetition",
                score=float(e - s),
            )


@TaggerRegistry.add("paragraph_tokenizer_repetitions_v1")
class ParagraphTokenizerRepetitionsTagger(TokenizerRepetitionsTagger):
    """Tagger to detect repetitions of tokens in paragraphs.
    It's faster than the tokenizer repetition tagger, but it does not account for
    repetitions of tokens that span multiple paragraphs."""

    def _extract_from_doc(self, doc: Document) -> Generator[Span, None, None]:
        offset = 0
        for paragraph in split_paragraphs(doc.text, remove_empty=False):
            # space is required to avoid first symbol in the paragraph to be
            # tokenized as a different token.
            for span in self._extract_from_text(" " + paragraph.text):
                span.start += offset - 1
                span.end += offset - 1
                yield span
            offset += len(paragraph.text)



import unittest

from dolma.core.data_types import Document
from dolma.taggers.repetitions import (
    ParagraphRepetitionsTagger,
    RepetitionsTagger,
    TokenizerRepetitionsTagger,
)

DOCUMENT_WITH_REPETITIONS = """
This is a text with repetitions.
This is a text with repetitions repetitions.
This is a text with repetitions repetitions repetitions.
This is a text with repetitions repetitions repetitions repetitions.

blah blah blah blah blah

No reps at the beginning of this sentence but MMMMMMMMMM

Seeing doubles: bass banana bass banana bass banana bass banana
"""

D0M0 = " repetitions repetitions repetitions repetitions"
D0M1 = "blah blah blah blah blah"
D0M2 = "MMMMMMMMMM"
D0M3 = " bass banana bass banana bass banana bass banana"

DOCUMENT_WITHOUT_REPETITIONS = """
This is a text without repetitions.

Absolutely no repetitions here.

There are double letters here and there but those don't count.
"""


class TestRepetitionsTagger(unittest.TestCase):
    def setUp(self) -> None:
        self.doc_with_reps = Document(source=__file__, id="0", text=DOCUMENT_WITH_REPETITIONS)
        self.doc_without_reps = Document(source=__file__, id="1", text=DOCUMENT_WITHOUT_REPETITIONS)

        self.repetitions_tagger = RepetitionsTagger()
        self.para_reps_tagger = ParagraphRepetitionsTagger()

        return super().setUp()

    def test_doc_with_repetitions(self):
        all_result = self.repetitions_tagger.predict(self.doc_with_reps)
        par_result = self.para_reps_tagger.predict(self.doc_with_reps)
        self.assertEqual(len(all_result.spans), 7)
        self.assertEqual(len(par_result.spans), 7)

        self.assertEqual(all_result.spans[0].type, "repetition")
        self.assertEqual(all_result.spans[0].select(self.doc_with_reps), D0M0)
        self.assertEqual(all_result.spans[0].score, len(D0M0))
        self.assertEqual(all_result.spans[0], par_result.spans[0])

        self.assertEqual(all_result.spans[1].type, "repetition")
        self.assertEqual(all_result.spans[1].select(self.doc_with_reps), D0M1)
        self.assertEqual(all_result.spans[1].score, len(D0M1))
        self.assertEqual(all_result.spans[1], par_result.spans[1])

        self.assertEqual(all_result.spans[2].type, "repetition")
        self.assertEqual(all_result.spans[2].select(self.doc_with_reps), D0M2)
        self.assertEqual(all_result.spans[2].score, len(D0M2))
        self.assertEqual(all_result.spans[2], par_result.spans[2])

        self.assertEqual(all_result.spans[3].type, "repetition")
        self.assertEqual(all_result.spans[3].select(self.doc_with_reps), D0M3)
        self.assertEqual(all_result.spans[3].score, len(D0M3))
        self.assertEqual(all_result.spans[3], par_result.spans[3])

        self.assertEqual(all_result.spans[4].type, "doc_max_repetition")
        self.assertEqual(all_result.spans[4].score, len(D0M0))
        self.assertEqual(all_result.spans[4], par_result.spans[4])

        matches_length = len(D0M0) + len(D0M1) + len(D0M2) + len(D0M3)
        self.assertEqual(all_result.spans[5].type, "doc_mean_repetition")
        self.assertEqual(all_result.spans[5].score, matches_length / 4)
        self.assertEqual(all_result.spans[5], par_result.spans[5])

        self.assertEqual(all_result.spans[6].type, "doc_frac_repetition")
        self.assertEqual(all_result.spans[6].score, matches_length / len(self.doc_with_reps.text))
        self.assertEqual(all_result.spans[6], par_result.spans[6])

    def test_doc_without_repetitions(self):
        all_result = self.repetitions_tagger.predict(self.doc_without_reps)
        par_result = self.para_reps_tagger.predict(self.doc_without_reps)
        self.assertEqual(len(all_result.spans), 3)
        self.assertEqual(len(par_result.spans), 3)

        self.assertEqual(all_result.spans[0].type, "doc_max_repetition")
        self.assertEqual(all_result.spans[0].score, 0)
        self.assertEqual(all_result.spans[0], par_result.spans[0])

        self.assertEqual(all_result.spans[1].type, "doc_mean_repetition")
        self.assertEqual(all_result.spans[1].score, 0)
        self.assertEqual(all_result.spans[1], par_result.spans[1])

        self.assertEqual(all_result.spans[2].type, "doc_frac_repetition")
        self.assertEqual(all_result.spans[2].score, 0)
        self.assertEqual(all_result.spans[2], par_result.spans[2])


class TestTokenizerRepetitionsTagger(unittest.TestCase):
    def setUp(self) -> None:
        self.doc_with_reps = Document(source=__file__, id="0", text=DOCUMENT_WITH_REPETITIONS)
        self.doc_without_reps = Document(source=__file__, id="1", text=DOCUMENT_WITHOUT_REPETITIONS)
        self.repetitions_tagger = TokenizerRepetitionsTagger()

        return super().setUp()

    def test_doc_with_repetitions(self):
        repeated_strings = [
            "repetitions repetitions repetitions",
            "repetitions repetitions repetitions repetitions",
            "blah blah blah blah",  # missing a blah bc the first element in this seq has diff token id
            "MMMMMMMM",  # shorter bc sequence is tokenized as 'ĠM', 'MM', 'MM', 'MM', 'MM', 'M'
            "bass banana bass banana bass banana bass banana",
        ]

        all_results = self.repetitions_tagger.predict(self.doc_with_reps)
        self.assertEqual(len(all_results.spans), len(repeated_strings) + 3)

        i = 0
        for string in repeated_strings:
            self.assertEqual(all_results.spans[i].type, "repetition")
            self.assertEqual(string, all_results.spans[i].select(self.doc_with_reps))
            self.assertEqual(len(string), all_results.spans[i].score)
            i += 1

        self.assertEqual(all_results.spans[i].type, "doc_max_repetition")
        self.assertEqual(all_results.spans[i].score, max(len(s) for s in repeated_strings))
        i += 1

        matches_length = sum(map(len, repeated_strings))
        self.assertEqual(all_results.spans[i].type, "doc_mean_repetition")
        self.assertEqual(all_results.spans[i].score, matches_length / len(repeated_strings))
        i += 1

        self.assertEqual(all_results.spans[i].type, "doc_frac_repetition")
        self.assertEqual(all_results.spans[i].score, matches_length / len(self.doc_with_reps.text))



import json
import argparse
import datetime
import apache_beam as beam
from apache_beam.io import Read
from apache_beam.io.textio import WriteToText, ReadFromText
from apache_beam.io.filesystem import CompressionTypes


DATA_ACQUISITION_DATE = '2023-03-16T01:43:55.831260+00:00'


def positive_int(value):
    value = int(value)
    if value <= 0:
        raise argparse.ArgumentTypeError(
            f'Value must be positive, {value} was passed.')
    return value


def build_base_parser():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--banned_subreddits_file",
        required=False,
        help="a text file containing a list of banned subreddits.",
        default="subreddit_blocklist.txt"
    )
    parser.add_argument(
        "--input_gcs_dir",
        required=False,
        help="Input google storage directory to read the data files from."
    )
    parser.add_argument(
        "--output_dir",
        required=False,
        help="Google cloud storage output directory to write the dataset.",
    )
    parser.add_argument(
        "--num_shards",
        default=750,
        type=positive_int,
        help="The number of shards in the dataset.",
    )
    return parser


def normalize_id(raw_id):
    """Reddit comment ids start with t1_, t2_, etc. which need to be stripped."""
    if raw_id:
        return raw_id[3:]
    return None


def normalize_string(text):
    def _safe_str(obj):
        try:
            return str(obj)
        except (UnicodeEncodeError, TypeError):
            return "[UNICODE ENCODE ERROR]"
    text = text.strip()
    return _safe_str(text)


def safe_load_json(line):
    try:
        return json.loads(line)
    except ValueError:
        return None


def isodatetime_from_epoch(epoch_time):
    try:
        return datetime.datetime.fromtimestamp(int(epoch_time), datetime.timezone.utc).isoformat()
    except TypeError:
        return epoch_time


def shuffle_posts(post_collection):
    post_collection |= "add random key" >> beam.Map(
        lambda value: (value['id'], value))
    post_collection |= "group by key" >> beam.GroupByKey()
    post_collection |= "get shuffled values" >> beam.FlatMap(lambda t: t[1])
    return post_collection


def convert_to_lm_format(example):
    text_field = 'conversational_format' if 'conversational_format' in example else 'formatted_post'

    formatted_data = {
        'text': example[text_field],
        'source': 'reddit',
        'created': example['created'],
        'added': example['added'],
        'id': example['id'],
        'metadata': {},
    }
    for k in list(example.keys()):
        if k not in [text_field] + list(formatted_data.keys()):
            formatted_data['metadata'][k] = example[k]

    formatted_data = json.dumps(formatted_data)
    return formatted_data


def load_filtered_subreddit_lists(blocked_subreddits_file):
    blocked_subreddits = [
        line.lower().strip() for line in open(
            blocked_subreddits_file,
            'r').readlines()]
    blocked_subreddits = list(set(blocked_subreddits))
    return blocked_subreddits


def read_content_from_source(content, p, args):
    if content is not None:
        content = p | ("Read in-memory content") >> beam.Create(content)
    else:
        content = p | ("Reading " + args.input_gcs_dir) >> Read(
            ReadFromText(args.input_gcs_dir)
        )
        content = content | (
            "Parse JSON" >> beam.Map(safe_load_json)
        )
    return content


def write_to_gcs(examples, banned_subreddits, args):
    examples = shuffle_posts(examples)
    examples = examples | ("Filter examples from banned subreddits" >> beam.Filter(
        lambda example: example['subreddit'].lower() not in banned_subreddits))

    file_name_suffix = ".jsonl.gz"
    serialize_fn = convert_to_lm_format
    write_sink = WriteToText
    compression_type = CompressionTypes.GZIP

    output_dir = args.output_dir
    name = 'sharded_output'
    serialized_examples = examples | (
        f'serialize {name} examples' >> beam.Map(serialize_fn))
    (
        serialized_examples | ("write " + name)
        >> write_sink(
            f'{output_dir}/{name}',
            file_name_suffix=file_name_suffix,
            compression_type=compression_type,
            max_bytes_per_shard=1000000000
        )
    )



"""

Filters.

@soldni

"""

from tokenizers import Tokenizer

from ..core.data_types import DocResult, Document, Span
from ..core.registry import TaggerRegistry
from ..core.taggers import BaseTagger


class BaseTokenizer(BaseTagger):
    TOKENIZER_PATH: str

    def __init__(self) -> None:
        if not hasattr(self, "TOKENIZER_PATH"):
            raise ValueError("TOKENIZER_PATH must be defined in the subclass")

        self.tokenizer = Tokenizer.from_pretrained(self.TOKENIZER_PATH)

    def predict(self, doc: Document) -> DocResult:
        tokens = self.tokenizer.encode(sequence=doc.text, add_special_tokens=False)
        return DocResult(doc=doc, spans=[Span(start=0, end=len(doc.text), type="tokens", score=len(tokens))])


@TaggerRegistry.add("tokenizers_EleutherAI_GPT_NeoX_20B")
class GPTNeoX20BTokenizer(BaseTokenizer):
    TOKENIZER_PATH = "EleutherAI/gpt-neox-20B"


@TaggerRegistry.add("tokenizers_AI2_OLMo_v1")
class OLMoV1Tokenizer(BaseTokenizer):
    TOKENIZER_PATH = "allenai/eleuther-ai-gpt-neox-20b-pii-special"



import random
from multiprocessing import current_process

from ..core.data_types import DocResult, Document, Span
from ..core.registry import TaggerRegistry
from ..core.taggers import BaseTagger


@TaggerRegistry.add("random_number_v1")
class RandomNumberTagger(BaseTagger):
    def __init__(self, seed: int = 1) -> None:
        assert seed > 0
        # we multiply the seed by the current process id to ensure that each
        # process has a different seed
        self.seed = ((current_process().pid or 0) + 1) * seed
        random.seed(self.seed)

    def predict(self, doc: Document) -> DocResult:
        score = random.random()
        return DocResult(doc=doc, spans=[Span(start=0, end=len(doc.text), type="random", score=score)])



"""

Tests for the utils module.

@kylel

"""


from unittest import TestCase

from dolma.core.data_types import TextSlice
from dolma.core.utils import split_paragraphs, split_sentences


class TestUtils(TestCase):
    def test_make_variable_name(self):
        pass

    def test_split_paragraphs(self):
        text = "This is a paragraph.\nThis is another paragraph.\nThis is a third paragraph."
        paragraphs = split_paragraphs(text=text)
        self.assertIsInstance(paragraphs[0], TextSlice)
        self.assertEqual(len(paragraphs), 3)
        self.assertEqual(paragraphs[0].text, "This is a paragraph.\n")
        self.assertEqual(text[paragraphs[0].start : paragraphs[0].end], paragraphs[0].text)
        self.assertEqual(paragraphs[1].text, "This is another paragraph.\n")
        self.assertEqual(text[paragraphs[1].start : paragraphs[1].end], paragraphs[1].text)

        paragraphs2 = split_paragraphs(text=text, remove_empty=False)
        self.assertListEqual([p.text for p in paragraphs], [p.text for p in paragraphs2])

    def test_split_paragraphs_empty(self):
        text = ""
        paragraphs = split_paragraphs(text=text)
        self.assertEqual(len(paragraphs), 0)

    def test_split_paragraphs_with_newline_and_spaces(self):
        text = "This is a sentence. \n  \n This is another sentence.\n\n  This is a third sentence."

        paragraphs = split_paragraphs(text=text)
        self.assertEqual(len(paragraphs), 3)
        self.assertIsInstance(paragraphs[0], TextSlice)
        self.assertEqual(len(paragraphs), 3)
        self.assertEqual(paragraphs[0].text, "This is a sentence. \n")
        self.assertEqual(text[paragraphs[0].start : paragraphs[0].end], paragraphs[0].text)
        self.assertEqual(paragraphs[1].text, " This is another sentence.\n")
        self.assertEqual(text[paragraphs[1].start : paragraphs[1].end], paragraphs[1].text)

        paragraphs = split_paragraphs(text=text, remove_empty=False)
        self.assertEqual(len(paragraphs), 5)
        self.assertIsInstance(paragraphs[0], TextSlice)
        self.assertEqual(len(paragraphs), 5)
        self.assertEqual(paragraphs[0].text, "This is a sentence. \n")
        self.assertEqual(text[paragraphs[0].start : paragraphs[0].end], paragraphs[0].text)
        self.assertEqual(paragraphs[1].text, "  \n")
        self.assertEqual(text[paragraphs[1].start : paragraphs[1].end], paragraphs[1].text)
        self.assertEqual(paragraphs[2].text, " This is another sentence.\n")
        self.assertEqual(text[paragraphs[2].start : paragraphs[2].end], paragraphs[2].text)

    def test_split_sentences(self):
        text = "This is a sentence. This is another sentence. This is a third sentence."

        sentences = split_sentences(text=text)
        self.assertIsInstance(sentences[0], TextSlice)
        self.assertEqual(len(sentences), 3)
        self.assertEqual(sentences[0].text, "This is a sentence.")
        self.assertEqual(text[sentences[0].start : sentences[0].end], sentences[0].text)
        self.assertEqual(sentences[1].text, "This is another sentence.")
        self.assertEqual(text[sentences[1].start : sentences[1].end], sentences[1].text)

    def test_split_sentences_empty(self):
        text = ""
        sentences = split_sentences(text=text)
        self.assertEqual(len(sentences), 0)

    def test_split_sentences_with_newline_and_spaces(self):
        text = "This is a sentence. \n  \n This is another sentence.\n\n  This is a third sentence."

        sentences = split_sentences(text=text)
        self.assertEqual(len(sentences), 3)
        self.assertIsInstance(sentences[0], TextSlice)
        self.assertEqual(len(sentences), 3)
        self.assertEqual(sentences[0].text, "This is a sentence.")
        self.assertEqual(text[sentences[0].start : sentences[0].end], sentences[0].text)
        self.assertEqual(sentences[1].text, "This is another sentence.")
        self.assertEqual(text[sentences[1].start : sentences[1].end], sentences[1].text)



import os

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

COLNAME_TO_LABEL = {
    "gopher_spans": "Gopher Rules",
    "decontamination_spans": "Decontamination",
    "hatespeech_spans": "Hate Speech",
    "pii_spans": "PII",
    "dedupe_paragraphs_spans": "Deduplication",
}


if os.path.exists("corr.csv"):
    corr = pd.read_csv("corr.csv", index_col=0)
else:
    # A line is e.g. 
    # {"gopher_span": [], "decontamination_span": [], "hatespeech_span": [], "pii_span": [], "dedupe_paragraphs_span": [[0, 615, 1.0], [615, 1214, 1.0], [1214, 1853, 1.0], [1853, 2417, 1.0], [2417, 2849, 1.0]]}
    df = pd.read_json(
        #"/home/niklas/dolma/tmp.jsonl/cc_en_head-0000.json", lines=True
        "cc_en_head_stats10.jsonl", lines=True
    )
    ### Matching based on the entire doc ###
    # Where the span is not empty turn it into True, elsewhere into False
    # Compute correlations between the attributes to later turn it into a heatmap
    corr = df.map(lambda x: bool(x)).corr(method='pearson')

    ### Matching based on spans ###
    """
    matrix = np.zeros((len(df.columns), len(df.columns)))
    columns = df.columns
    for _, row in df.iterrows():
        # Iterate over the columns
        for i, col1 in enumerate(columns):
            for j, col2 in enumerate(columns):
                # If the columns are the same, skip
                if col1 == col2: continue
                # Increment if the spans overlap
                # e.g. [0, 615, 1.0] & [614, 1214, 1.0] -> 1
                # while [0, 615, 1.0] & [700, 1214, 1.0] -> 0
                matrix[i, j] += float(
                    any(
                        [span1[0] <= span2[0] and span1[1] >= span2[0] for span2 in row[col2]]
                        for span1 in row[col1]
                    )
                )

    corr = matrix / len(df)
    corr *= 100
    # Add the column names
    corr = pd.DataFrame(corr, columns=columns, index=columns)                
    """

# Plot the heatmap
plt.figure(figsize=(36, 24))
# define the mask to set the values in the upper triangle to True
mask = np.triu(np.ones_like(corr, dtype=bool))
heatmap = sns.heatmap(
    corr.rename(columns=COLNAME_TO_LABEL, index=COLNAME_TO_LABEL),
    mask=mask, 
    vmin=corr.values.min(), 
    vmax=corr.values[~mask].max(), # Max ignoring the ones in corr
    annot=True, 
    cmap='Blues',
    linewidths=0.5,
    annot_kws={"fontsize": 32},
    cbar=False, # No legend
)

heatmap.set_xticklabels(heatmap.get_xmajorticklabels(), fontsize=32)#, fontweight="bold")
heatmap.set_yticklabels(heatmap.get_ymajorticklabels(), fontsize=32)#, fontweight="bold")

corr.to_csv("corr.csv")
plt.savefig('attributes_heatmap_docbased_9mdocs.pdf', dpi=450, bbox_inches='tight')
plt.savefig('attributes_heatmap_docbased_9mdocs.png', dpi=450, bbox_inches='tight')



import importlib
import os
import re
import string
import sys
from typing import List, Union

try:
    import blingfire

    BLINGFIRE_AVAILABLE = True
except Exception:
    BLINGFIRE_AVAILABLE = False

import nltk
import uniseg.wordbreak
from nltk.tokenize.punkt import PunktSentenceTokenizer

try:
    nltk.data.find("tokenizers/punkt")
except LookupError:
    nltk.download("punkt")


from .data_types import TextSlice
from .loggers import get_logger

sent_tokenizer = PunktSentenceTokenizer()
logger = get_logger(__name__)


def make_variable_name(name: str, remove_multiple_underscores: bool = False) -> str:
    # use underscores for any non-valid characters in variable name
    name = re.sub(r"[^a-zA-Z0-9_]", "_", name)

    if remove_multiple_underscores:
        # replace multiple underscores with a single underscore
        name = re.sub(r"__+", "_", name)

    if name[0] in string.digits:
        raise ValueError(f"Invalid variable name {name}")

    return name


def split_words(text: str, remove_empty: bool = True) -> List[TextSlice]:
    """
    Split a string into words, as defined by the unicode standard.
    For more info, see https://www.unicode.org/reports/tr29/
    """
    text_slices: List[TextSlice] = []
    offset = 0
    for word in uniseg.wordbreak.words(text):
        if word.strip() or not remove_empty:
            text_slices.append(TextSlice(doc=text, start=offset, end=offset + len(word)))
        offset += len(word)
    return text_slices


def split_paragraphs(text: str, remove_empty: bool = True) -> List[TextSlice]:
    """
    Split a string into paragraphs. A paragraph is defined as a sequence of zero or more characters, followed
    by a newline character, or a sequence of one or more characters, followed by the end of the string.

    Args:
        text (str): The text to split into paragraphs.
        remove_empty (bool): Whether to remove empty paragraphs. Defaults to True.
    """
    text_slices = [
        TextSlice(doc=text, start=match.start(), end=match.end())
        for match in re.finditer(r"([^\n]*\n|[^\n]+$)", text)
    ]
    if remove_empty:
        text_slices = [text_slice for text_slice in text_slices if text_slice.text.strip()]
    return text_slices


def split_sentences(text: str, remove_empty: bool = True) -> List[TextSlice]:
    """
    Split a string into sentences.
    """
    if text and BLINGFIRE_AVAILABLE:
        _, offsets = blingfire.text_to_sentences_and_offsets(text)  # pyright: ignore
    elif text:
        offsets = [(start, end) for start, end in sent_tokenizer.span_tokenize(text)]
    else:
        offsets = []

    if remove_empty is True:
        return [TextSlice(doc=text, start=start, end=end) for (start, end) in offsets]
    else:
        raise NotImplementedError("remove_empty=False is not implemented yet")


def import_modules(modules_path: Union[List[str], None]):
    """Import one or more user modules from either names or paths.
    Importing from path is modeled after fairseq's import_user_module function:
    github.com/facebookresearch/fairseq/blob/da8fb630880d529ab47e53381c30ddc8ad235216/fairseq/utils.py#L464

    Args:
        modules_path (Union[List[str], None]): List of module names or paths to import.
    """

    for module_path in modules_path or []:
        # try importing the module directly
        try:
            importlib.import_module(module_path)
            continue
        except ModuleNotFoundError:
            pass
        except Exception as exp:
            raise RuntimeError(f"Could not import module {module_path}: {exp}") from exp

        # if that fails, try importing the module as a path

        # check if this function has a memorization attribute; if not; create it
        # the memorization attribute is used to ensure that user modules are only imported once
        if (modules_memo := getattr(import_modules, "memo", None)) is None:
            modules_memo = set()
            import_modules.memo = modules_memo  # type: ignore

        # ensure that user modules are only imported once
        if module_path not in modules_memo:
            modules_memo.add(module_path)

            if not os.path.exists(module_path):
                raise FileNotFoundError(f"Could not find module {module_path}")

            # the format is `<module_parent>/<module_name>.py` or `<module_parent>/<module_name>`
            module_parent, module_name = os.path.split(module_path)
            module_name, _ = os.path.splitext(module_name)
            if module_name not in sys.modules:
                sys.path.insert(0, module_parent)
                importlib.import_module(module_name)
            elif module_path in sys.modules[module_name].__path__:
                logger.info(f"{module_path} has already been imported.")
            else:
                raise ImportError(
                    f"Failed to import {module_path} because the corresponding module name "
                    f"({module_name}) is not globally unique. Please rename the directory to "
                    "something unique and try again."
                )



from contextlib import ExitStack
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional

import smart_open
from omegaconf import OmegaConf as om

from dolma import deduper
from dolma.cli import BaseCli, field, print_config
from dolma.cli.shared import WorkDirConfig, get_path_to_temp_file, make_workdirs
from dolma.core.errors import DolmaConfigError
from dolma.core.loggers import get_logger
from dolma.core.paths import glob_path, is_local


@dataclass
class ParagraphDedupeConfig:
    attribute_name: Optional[str] = field(help="Name of the output field in the tagger")


@dataclass
class DocumentDedupeConfig:
    attribute_name: Optional[str] = field(help="Name of the output field in the tagger")
    key: str = field(help="Name of the input field to use for deduplication, e.g. `$.metadata.url`")


@dataclass
class BloomFilterConfig:
    file: str = field(help="Path where to read/write the bloom filter file to/from. Required.")
    size_in_bytes: int = field(
        default=0,
        help=(
            "Size of the bloom filter in bytes. Either this value is provided, or both estimated_doc_count "
            "and desired_false_positive_rate."
        ),
    )
    read_only: bool = field(help="If true, the bloom filter will be read from the file and not updated. Required.")
    estimated_doc_count: int = field(
        default=0,
        help=(
            "Estimated number of documents to be added to the bloom filter. Either this value is provided, "
            "or both size_in_bytes and desired_false_positive_rate."
        ),
    )
    desired_false_positive_rate: float = field(
        default=0,
        help=(
            "Desired false positive rate. Either this value is provided, or both size_in_bytes and "
            "estimated_doc_count."
        ),
    )


@dataclass
class DedupeConfig:
    name: str = field(help="Name of the deduper. Required.")
    documents: Optional[DocumentDedupeConfig] = field(
        default=None, help="Configuration for document deduplication"
    )
    paragraphs: Optional[ParagraphDedupeConfig] = field(
        default=None, help="Configuration for paragraph deduplication"
    )
    skip_empty: Optional[bool] = field(default=False, help="If true, empty documents/paragraphs will be skipped")


@dataclass
class DeduperConfig:
    documents: List[str] = field(default=[], help="Paths to the documents to be deduplicated. Required.")
    work_dir: WorkDirConfig = field(default=WorkDirConfig(), help="Configuration for temporary work directories.")
    dedupe: DedupeConfig = field(help="Deduplication configuration. Required.")
    bloom_filter: BloomFilterConfig = field(help="Bloom filter configuration. Required.")
    processes: int = field(
        default=1, help="Number of processes to use for deduplication. If 1, no multiprocessing will be used."
    )
    dryrun: bool = field(
        default=False,
        help="If true, only print the configuration and exit without running the deduper.",
    )


class DeduperCli(BaseCli):
    CONFIG = DeduperConfig
    DESCRIPTION = "Deduplicate documents or paragraphs using a bloom filter."

    @classmethod
    def run(cls, parsed_config: DeduperConfig):
        logger = get_logger("tagger")

        dict_config: Dict[str, Any] = {}

        with ExitStack() as stack:
            work_dirs = stack.enter_context(make_workdirs(parsed_config.work_dir))

            # create a dedupe config to populate
            dedupe_dict_config: Dict[str, Any] = {"skip_empty": parsed_config.dedupe.skip_empty}
            try_name = parsed_config.dedupe.name if not om.is_missing(parsed_config.dedupe, "name") else None

            # add either the document or paragraph dedupe config
            if not (
                om.is_missing(parsed_config.dedupe.documents, "attribute_name")
                and om.is_missing(parsed_config.dedupe.documents, "key")
            ):
                cfg = om.to_container(parsed_config.dedupe.documents)
                assert isinstance(cfg, dict), "Expected dedupe.documents to be a dict"
                dedupe_dict_config["documents"] = cfg
                try_name = try_name or cfg["attribute_name"]
            elif not om.is_missing(parsed_config.dedupe.paragraphs, "attribute_name"):
                cfg = om.to_container(parsed_config.dedupe.paragraphs)
                assert isinstance(cfg, dict), "Expected dedupe.paragraphs to be a dict"
                dedupe_dict_config["paragraphs"] = cfg
                try_name = try_name or cfg["attribute_name"]
            else:
                raise ValueError("Either dedupe.documents or dedupe.paragraphs must be specified")

            if try_name is None:
                raise ValueError("dedupe.name must be specified")
            dedupe_dict_config["name"] = try_name

            # add the dedupe config to the main config
            dict_config["dedupe"] = dedupe_dict_config

            # perform some path validation to make sure we don't call the mixer with invalid config
            total_matching_documents = 0
            for document in parsed_config.documents:
                dict_config.setdefault("documents", []).append(str(document))

                if document.count("*") > 1:
                    raise DolmaConfigError("Only one wildcard is allowed in the document path")

                current_matching_documents = sum(1 for _ in glob_path(document))
                if current_matching_documents == 0:
                    # only raise a warning if no documents are found for a single path
                    logger.warning("No documents found for path %s", document)
                total_matching_documents += current_matching_documents

            if total_matching_documents == 0:
                # but raise an error if no documents are found for all paths
                raise DolmaConfigError(f"No documents found for the paths {dict_config['documents']}.")

            # The rust deduper does not work with remote files, so we need to download the bloom filter
            # if it is not local. If the remote file does not exists, and the bloom filter is read-only,
            # we raise an error.
            if not (path_is_local := is_local(parsed_config.bloom_filter.file)):
                local_bloom_file = stack.enter_context(get_path_to_temp_file())
                try:
                    with smart_open.open(parsed_config.bloom_filter.file, "rb") as f:
                        contents: bytes = f.read()  # pyright: ignore
                    local_bloom_file.write_bytes(contents)
                except (FileNotFoundError, OSError) as ex:
                    if parsed_config.bloom_filter.read_only:
                        raise ex
            else:
                local_bloom_file = Path(parsed_config.bloom_filter.file)

            dict_config["bloom_filter"] = {
                "file": str(local_bloom_file),
                "read_only": bool(parsed_config.bloom_filter.read_only),
                "size_in_bytes": int(parsed_config.bloom_filter.size_in_bytes),
                "estimated_doc_count": int(parsed_config.bloom_filter.estimated_doc_count),
                "desired_false_positive_rate": float(parsed_config.bloom_filter.desired_false_positive_rate),
            }

            if dict_config["bloom_filter"]["size_in_bytes"] <= 0 and (
                dict_config["bloom_filter"]["estimated_doc_count"] <= 0
                or dict_config["bloom_filter"]["desired_false_positive_rate"] <= 0
            ):
                raise ValueError(
                    "Either bloom_filter.size_in_bytes or bloom_filter.estimated_doc_count and "
                    "bloom_filter.desired_false_positive_rate must be specified"
                )

            dict_config["work_dir"] = {"input": str(work_dirs.input), "output": str(work_dirs.output)}
            dict_config["processes"] = int(parsed_config.processes)

            if len(dict_config["documents"]) == 0:
                raise ValueError("At least one document must be specified")

            print_config(dict_config)
            if parsed_config.dryrun:
                logger.info("Exiting due to dryrun.")
                return

            # run the deduper
            deduper(dict_config)

            # upload to remote file if necessary
            if not parsed_config.bloom_filter.read_only and not path_is_local:
                print(f"Pushing Bloom filter to {parsed_config.bloom_filter.file}")
                local = stack.enter_context(smart_open.open(local_bloom_file, "rb"))
                remote = stack.enter_context(smart_open.open(parsed_config.bloom_filter.file, "wb"))
                remote.write(local.read())



"""
Find source document given `dolma tokens` metadata and offset.

@soldni
"""


import argparse
import json
from bisect import bisect
from typing import Any, Dict, List

import pandas as pd
import smart_open
import tqdm


def parse_args() -> argparse.Namespace:
    ap = argparse.ArgumentParser()
    ap.add_argument("-f", "--file", type=str, required=True)
    ap.add_argument("-o", "--offsets", type=int, nargs="+", required=True)
    ap.add_argument("-d", "--destination", type=str, required=True)
    return ap.parse_args()


def main() -> None:
    opts = parse_args()

    print("Loading metadata...", end=" ", flush=True)
    data = pd.read_csv(opts.file, header=0, names=["start", "end", "sha", "path", "pos"])

    locs: List[int] = []
    for offset in opts.offsets:
        loc = bisect(data["start"].values, offset)
        if data["start"][loc] > offset:
            # in case of not exact match
            loc -= 1

        if data.iloc[loc].start >= offset or data.iloc[loc].end <= offset:
            raise ValueError(f"Offset {offset} not found in {opts.file}.")
        locs.append(loc)

    print(f"offset(s) found at {locs} row(s)", flush=True)

    rows = data.iloc[locs]

    documents: List[Dict[str, Any]] = []
    for i, row in rows.iterrows():
        with smart_open.open(row.path) as f:
            for i, ln in tqdm.tqdm(enumerate(f, start=1), desc="Searching doc...", unit="lines", total=row.pos):
                if i == row.pos:
                    documents.append(json.loads(ln))
                    break
                elif i > row.pos:
                    raise ValueError("location in document not found")

    with smart_open.open(opts.destination, "w") as f:
        for doc in documents:
            f.write(json.dumps(doc) + "\n")


if __name__ == "__main__":
    main()



import unittest

from dolma.core.data_types import Document
from dolma.taggers.length import CharLengthStripWsV1, CharLengthV1


class TestLengthTaggers(unittest.TestCase):
    def setUp(self) -> None:
        self.doc_with_ws = Document(text="\n\tThis is a test.  ", id="0", source=__file__)
        self.doc_no_ws = Document(text="This is a test.", id="1", source=__file__)
        self.doc_has_inter_ws = Document(text="This is\n a test.", id="2", source=__file__)
        self.empty_doc = Document(text="", id="3", source=__file__)
        self.empty_doc_with_ws = Document(text="\n\t  \n", id="4", source=__file__)

        self.char_length_v1 = CharLengthV1()
        self.char_length_strip_ws_v1 = CharLengthStripWsV1()

    def test_doc_with_no_ws(self):
        result = self.char_length_v1.predict(self.doc_no_ws)
        self.assertEqual(result.spans[0].score, 15)

        result = self.char_length_strip_ws_v1.predict(self.doc_no_ws)
        self.assertEqual(result.spans[0].score, 15)

    def test_doc_with_ws(self):
        result = self.char_length_v1.predict(self.doc_with_ws)
        self.assertEqual(result.spans[0].score, 19)

        result = self.char_length_strip_ws_v1.predict(self.doc_with_ws)
        self.assertEqual(result.spans[0].score, 15)

    def test_doc_with_inter_ws(self):
        result = self.char_length_v1.predict(self.doc_has_inter_ws)
        self.assertEqual(result.spans[0].score, 16)

        result = self.char_length_strip_ws_v1.predict(self.doc_has_inter_ws)
        self.assertEqual(result.spans[0].score, 16)

    def test_empty_doc(self):
        result = self.char_length_v1.predict(self.empty_doc)
        self.assertEqual(result.spans[0].score, 0)

        result = self.char_length_strip_ws_v1.predict(self.empty_doc)
        self.assertEqual(result.spans[0].score, 0)

    def test_empty_doc_with_ws(self):
        result = self.char_length_v1.predict(self.empty_doc_with_ws)
        self.assertEqual(result.spans[0].score, 5)

        result = self.char_length_strip_ws_v1.predict(self.empty_doc_with_ws)
        self.assertEqual(result.spans[0].score, 0)



"""

Unit tests for code taggers.

@soldni

"""

import re
import unittest

from bs4 import BeautifulSoup

from dolma.core.data_types import Document, DocumentWithMetadata
from dolma.taggers.code import (
    CodeCopyrightTagger,
    CodeRedPajamaTaggers,
    CodeSecretsTagger,
    CodeStarCoderTaggers2,
)

DOC_WITH_SECRETS_AND_COPYRIGHT = """
/* copyright: Test 2023 **/

This is a document.

This line contains a secret: https://username:password@dolma.allen.ai

This is a line with just text.
"""


class TestCodeTaggers(unittest.TestCase):
    def setUp(self) -> None:
        self.doc = Document(id="0", text=DOC_WITH_SECRETS_AND_COPYRIGHT.strip(), source=__file__)
        return super().setUp()

    def test_code_secrets_tagger(self) -> None:
        tagger = CodeSecretsTagger()
        result = tagger.predict(self.doc)

        self.assertEqual(len(result.spans), 3)

        self.assertEqual(result.spans[0].type, "SECRET_Secret_Keyword")
        self.assertEqual(result.spans[0].select(self.doc), "https://username:password@dolma.allen.ai")

        self.assertEqual(result.spans[1].type, "SECRET_Basic_Auth_Credentials")
        self.assertEqual(result.spans[1].select(self.doc), "password")

        self.assertEqual(result.spans[2].type, "doc")
        self.assertEqual(result.spans[2].select(self.doc), self.doc.text)

    def test_copyright_notice(self):
        tagger = CodeCopyrightTagger()
        result = tagger.predict(self.doc)

        self.assertEqual(len(result.spans), 2)

        self.assertEqual(result.spans[0].type, "copyright_notice")
        self.assertEqual(result.spans[0].select(self.doc), "/* copyright: Test 2023 **/")

        self.assertEqual(result.spans[1].type, "doc")
        self.assertEqual(result.spans[1].select(self.doc), self.doc.text)


class TestRedPajamaTaggers(unittest.TestCase):
    def setUp(self) -> None:
        self.doc = Document(id="0", text=DOC_WITH_SECRETS_AND_COPYRIGHT.strip(), source=__file__)
        self.whitespace_regex = re.compile(r"\w+|[^\w\s]+")
        return super().setUp()

    def test_code_red_pajama_taggers(self) -> None:
        tagger = CodeRedPajamaTaggers()
        result = tagger.predict(self.doc)

        # handy precomputed values
        line_lengths = list(map(len, self.doc.text.splitlines()))
        words = self.whitespace_regex.findall(self.doc.text)
        self.assertGreater(len(line_lengths), 0)
        self.assertGreater(len(words), 0)

        self.assertEqual(len(result.spans), 4)
        self.assertEqual(result.spans[0].type, "max_line_length_doc")
        self.assertEqual(result.spans[0].score, max(line_lengths))

        self.assertEqual(result.spans[1].type, "avg_line_length_doc")
        self.assertEqual(result.spans[1].score, sum(line_lengths) / len(line_lengths))

        self.assertEqual(result.spans[2].type, "alnum_prop_doc")
        self.assertEqual(result.spans[2].score, len(list(filter(str.isalnum, self.doc.text))) / len(self.doc.text))

        # TODO: This test fail; check with Akshita if this is expected
        # self.assertEqual(result.spans[3].type, "alpha_token_prop_doc")
        # self.assertEqual(result.spans[3].score, len(list(filter(str.isalpha, words))) / len(words))


DOC_WITH_METADATA = """
An XML file begins as follows:

```
<?xml version="1.0" encoding="UTF-8"?>
...
</xml>
```

An HTML file begins as follows:

```
<!DOCTYPE html>
<html>
...
</html>
```

These are different.
"""

DOC_WITH_PYTHON_CONTENT = """
def foo():
    # prints hello world
    print("Hello, World!")
"""

DOC_WITH_HTML_CONTENT = """
<!DOCTYPE html>
<html>
<head>
    <title>Page Title</title>
</head>
<body>
    <h1>This is a Heading</h1>
    <p>This is a paragraph.</p>
</body>
<javascript>
    console.log("Hello, World!")
    for (let i = 0; i < 10; i++) {
        console.log(i)
    }
</javascript>
</html>
"""


class TestStarCoderTaggers(unittest.TestCase):
    def setUp(self) -> None:
        self.md_doc = DocumentWithMetadata(
            id="0",
            text=DOC_WITH_METADATA.strip(),
            source=__file__,
            metadata={"ext": "md", "max_stars_count": 10},
        )
        self.python_doc = DocumentWithMetadata(
            id="1",
            text=DOC_WITH_PYTHON_CONTENT.strip(),
            source=__file__,
            metadata={"ext": "py", "max_stars_count": 1},
        )
        self.html_doc = DocumentWithMetadata(
            id="2",
            text=DOC_WITH_HTML_CONTENT.strip(),
            source=__file__,
            metadata={"ext": "html", "max_stars_count": 5},
        )
        self.tagger = CodeStarCoderTaggers2()
        return super().setUp()

    def test_metadata_tagger(self):
        result = self.tagger.predict(self.md_doc)
        self.assertEqual(len(result.spans), 4)

        self.assertEqual(result.spans[0].type, "has_xml_template_doc")
        self.assertEqual(result.spans[0].score, 1.0)

        self.assertEqual(result.spans[1].type, "num_github_stars_doc")
        self.assertEqual(result.spans[1].score, 10.0)

        # not a python, js, or java, so this is pinned to 0.5
        self.assertEqual(result.spans[2].type, "code_to_comment_ratio_doc")
        self.assertEqual(result.spans[2].score, 0.5)

        # not html, so this is pinned to 1.0
        self.assertEqual(result.spans[3].type, "code_to_text_ratio_html_doc")
        self.assertEqual(result.spans[3].score, 1.0)

    def test_python_tagger(self):
        result = self.tagger.predict(self.python_doc)

        comment_lines = [
            lns.split("#")[1].strip()
            for ln in self.python_doc.text.split("\n")
            if (lns := ln.strip()).startswith("#")
        ]

        self.assertEqual(len(result.spans), 4)

        self.assertEqual(result.spans[0].type, "has_xml_template_doc")
        self.assertEqual(result.spans[0].score, 0.0)

        self.assertEqual(result.spans[1].type, "num_github_stars_doc")
        self.assertEqual(result.spans[1].score, 1.0)

        self.assertEqual(result.spans[2].type, "code_to_comment_ratio_doc")
        self.assertEqual(result.spans[2].score, sum(map(len, comment_lines)) / len(self.python_doc.text))

        self.assertEqual(result.spans[3].type, "code_to_text_ratio_html_doc")
        self.assertEqual(result.spans[3].score, 1.0)

    def test_html_tagger(self):
        result = self.tagger.predict(self.html_doc)

        soup = BeautifulSoup(self.html_doc.text, features="html.parser")

        self.assertEqual(len(result.spans), 4)

        self.assertEqual(result.spans[0].type, "has_xml_template_doc")
        self.assertEqual(result.spans[0].score, 0.0)

        self.assertEqual(result.spans[1].type, "num_github_stars_doc")
        self.assertEqual(result.spans[1].score, 5.0)

        self.assertEqual(result.spans[2].type, "code_to_comment_ratio_doc")
        self.assertEqual(result.spans[2].score, 0.5)

        self.assertEqual(result.spans[3].type, "code_to_text_ratio_html_doc")
        self.assertEqual(result.spans[3].score, len(soup.get_text()) / len(self.html_doc.text))

    def test_html_tagger_doc_too_short(self):
        doc = DocumentWithMetadata(
            id="3",
            text="<html><head></head><body></body></html>",
            source=__file__,
            metadata={"ext": "html", "max_stars_count": 5},
        )
        doc.text = doc.text[:100]
        result = self.tagger.predict(doc)

        self.assertEqual(result.spans[3].type, "code_to_text_ratio_html_doc")
        self.assertEqual(result.spans[3].score, 0.0)



"""A Dataflow script for creating DOLMA formatted pretraining data from
reddit comment dumps.

For usage see README.md.

Adapted from work done at PolyAI:
https://github.com/PolyAI-LDN/conversational-datasets/tree/master/reddit
"""


import logging
from functools import partial
import uuid
import random
from collections import defaultdict
import apache_beam as beam
from apache_beam.options.pipeline_options import PipelineOptions
from utils.shared_utils import (
    build_base_parser,
    positive_int,
    normalize_id,
    normalize_string,
    load_filtered_subreddit_lists,
    read_content_from_source,
    write_to_gcs,
    isodatetime_from_epoch,
    DATA_ACQUISITION_DATE
)


def parse_args(argv=None):
    parser = build_base_parser()

    parser.add_argument(
        "--parent_depth",
        type=positive_int,
        default=12,
        help="How many parents to consider.",
    )
    parser.add_argument(
        "--max_length",
        type=positive_int,
        default=2048,  # was 127 then 512.
        help="Maximum letters in content to include.",
    )
    parser.add_argument(
        "--min_length",
        type=positive_int,
        default=9,
        help="Minimum letters in content to include.",
    )
    return parser.parse_known_args(argv)


def normalize_comment(comment, max_length):

    def trim(text, max_length):
        """Trims text to be at most `max_length`, without splitting apart words."""
        if len(text) <= max_length:
            return text
        text = text[:max_length + 1]

        # Trim until the last two characters are the boundary between an
        # alphanumeric character, and a non-alphanumeric character.
        while len(text) > 1 and (text[-1].isalnum() == text[-2].isalnum()):
            text = text[:-1]
        return text[:-1]

    comment = {
        'id': comment['id'],
        'thread_id': normalize_id(comment['link_id']),
        'parent_id': normalize_id(comment['parent_id']),
        'body': trim(normalize_string(comment['body']), max_length),
        'body_is_trimmed': len(comment['body']) > max_length,
        'author': normalize_string(comment['author']),
        'subreddit': normalize_string(comment['subreddit']),
        'created_utc': comment['created_utc'],
        'score': comment['score'],
        'link_id': comment['link_id']
    }
    return comment


def create_examples(thread, parent_depth, min_length):
    def _should_skip(comment, min_length):
        if comment['body_is_trimmed']:
            return True
        if comment['body'] in {
            "[deleted]",
            "[removed]",
            "[UNICODE ENCODE ERROR]"}:
            return True
        if comment['subreddit'] in {
            "[deleted]",
            "[removed]",
            "[UNICODE ENCODE ERROR]"}:
            return True
        if len(comment['body']) < min_length:
            return True
        return False

    def linear_paths(id_to_comment, parent_depth):
        """Gets all linear paths of content and replies from the thread.
        Each linear path is guaranteed to contain at least two comments."""
        paths = []
        seen_ids = set()
        id_to_children = defaultdict(list)
        for comment_id, comment in list(id_to_comment.items()):
            id_to_children[comment['parent_id']].append(comment_id)
            if comment['parent_id'] not in id_to_comment:
                paths.append([comment_id])
                seen_ids.add(comment_id)

        # paths start with root content
        while paths:
            new_paths = []
            for path in paths:
                last_id = path[-1]
                for child_id in id_to_children[last_id]:
                    if child_id in seen_ids:
                        # Prevent infinite loops.
                        continue
                    seen_ids.add(child_id)
                    new_path = path[-parent_depth:] + [child_id]
                    new_paths.append(new_path)

            # yield all unique paths at the tallest depth
            if len(paths) > 0:
                for path in paths:
                    root_node = path[0]
                    last_node = path[-1]
                    if root_node not in [p[0] for p in new_paths]:
                        yield path
                    elif last_node not in [p[-2] for p in new_paths]:
                        yield path
            paths = new_paths

    def trim_paths(linear_paths, id_to_comment, min_length):
        clean_paths = []
        for p in linear_paths:
            clean_path = []
            for c in p:
                if not _should_skip(id_to_comment[c], min_length):
                    clean_path.append(c)
                else:
                    break
            if len(clean_path) >= 2:
                clean_paths.append(clean_path)
        return clean_paths

    def choose_longest_unique_path(linear_paths):
        longest_paths = []
        seen_comments = []
        for p in linear_paths[::-1]:
            if p[0] not in seen_comments:
                longest_paths.append(p)
                seen_comments.extend(p)
        return longest_paths

    def build_anonymous_author_names(thread):
        author_to_anonymous_name = {}
        for comment in list(thread):
            rand_int = random.randrange(1, 10 ** 4)
            random_author_name = f'user_{str(rand_int).zfill(4)}'
            author_to_anonymous_name[comment['author']] = random_author_name
        return author_to_anonymous_name

    id_to_comment = {
        comment['id']: comment for comment in list(thread)}

    anonymous_author_lookup = build_anonymous_author_names(thread)

    all_linear_paths = [p for p in linear_paths(id_to_comment, parent_depth) if len(p) >= 2]
    cleaned_paths = trim_paths(all_linear_paths, id_to_comment, min_length)
    longest_linear_paths = choose_longest_unique_path(cleaned_paths)

    for linear_path in longest_linear_paths:
        response = id_to_comment[linear_path[-1]]
        context = id_to_comment[linear_path[-2]]  # guaranteed to exist.

        example = {
            'subreddit': response['subreddit'],
            'thread_id': response['thread_id'],
            'created': isodatetime_from_epoch(response['created_utc']),
            'added': DATA_ACQUISITION_DATE,
            'id': uuid.uuid4().hex,
        }

        def _get_author_offsets(author, conv_string):
            return len(conv_string), len(author) + len(conv_string)

        conv_str = ''
        author_offsets = []
        comment_scores = []
        for i, parent_id in enumerate(linear_path[:-2]):
            if i >= parent_depth:
                break
            author = id_to_comment[parent_id]['author']
            author_offsets.append(_get_author_offsets(author, conv_str))
            comment_scores.append(id_to_comment[parent_id]['score'])
            conv_str += f"#{anonymous_author_lookup[author]}#: {id_to_comment[parent_id]['body']}\n\n"
        author_offsets.append(_get_author_offsets(context['author'], conv_str))
        conv_str += f"#{anonymous_author_lookup[context['author']]}#: {context['body']}"
        author_offsets.append(_get_author_offsets(response['author'], conv_str))
        conv_str += f"#{anonymous_author_lookup[response['author']]}#: {response['body']}"
        comment_scores.append(context['score'])
        comment_scores.append(response['score'])

        example['conversational_format'] = conv_str
        example['user_id_spans'] = author_offsets
        example['comment_scores'] = comment_scores

        yield example


def run(argv=None, comments=None):

    args, pipeline_args = parse_args(argv)
    banned_subreddits_file = args.banned_subreddits_file
    banned_subreddits = load_filtered_subreddit_lists(banned_subreddits_file)

    pipeline_options = PipelineOptions(pipeline_args, save_main_session=True,
                                       dataflow_service_options=['enable_prime'],
                                       experiments=['enable_batch_vmr',
                                                    'enable_vertical_memory_autoscaling',
                                                    'no_use_multiple_sdk_containers'])

    p = beam.Pipeline(options=pipeline_options)

    comments = read_content_from_source(comments, p, args)

    comments |= (
        "normalize content" >> beam.Map(
            partial(normalize_comment, max_length=args.max_length)))

    thread_id_to_comments = comments | (
        "Key by thread id" >> beam.Map(
            lambda comment: (comment['thread_id'], comment)))

    threads = thread_id_to_comments | (
        "Group content by thread ID" >> beam.GroupByKey())

    threads = threads | ("Get threads" >> beam.Map(lambda t: t[1]))

    examples = threads | (
        "Create examples" >> beam.FlatMap(
            partial(create_examples,
                    parent_depth=args.parent_depth,
                    min_length=args.min_length,
                    )))

    # examples = shuffle_posts(examples)

    write_to_gcs(examples, banned_subreddits, args)

    result = p.run()
    result.wait_until_finish()


if __name__ == "__main__":
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    run()



from typing import Generator, List, NamedTuple

import numpy as np


def find_end_first_consecutive_true(arr: np.ndarray) -> int:
    """Function to find the end position of the first consecutive sequence of True in an array."""
    if not arr[0]:
        return 0

    prog = np.cumsum(arr)
    if prog[-1] == len(arr):
        return len(arr)

    true_locs = np.where(prog[:-1:] == prog[1::])[0]

    return true_locs[0] + 1


def find_start_last_consecutive_true(arr: np.ndarray) -> int:
    """Function to find the start position of the last consecutive sequence of True in an array."""
    reverse = find_end_first_consecutive_true(arr[::-1])
    return len(arr) - reverse if reverse > 0 else -1


def group_consecutive_values(arr: np.ndarray, stepsize: int = 1) -> List[np.ndarray]:
    """Function to group consecutive values in an array."""
    return np.split(arr, np.where(np.diff(arr) != stepsize)[0] + 1)


class RepetitionTuple(NamedTuple):
    """Tuple to store information about a periodic sequence."""

    start: int
    end: int
    period: int
    times: int


def find_periodic_sequences(
    arr: np.ndarray, max_period: int, min_period: int = 1, mask_value: int = -1
) -> Generator[RepetitionTuple, None, None]:
    """Function to find periodic sequences in an array.

    This function sweeps through the array and checks for sequences of length
    [min_period, max_period] that repeat at least 3 times. To do so, it
    reshape the array into a matrix with `period` columns and checks if each
    row is equal to the previous row. Blocks of repeating rows indicates repeating
    sequences.

    Because there's no guarantee that the sequences start at the beginning of each
    row, it can only detect sequences that repeat at least 3 times. To account
    for the fact that sequences may not start at the beginning of each row (or
    end at the end of each row), we check the end of the previous row and the
    start of the next row to determine the actual start and end positions of the
    sequence.

    Args:
        arr (np.ndarray): The array to search for periodic sequences.
        max_period (int): The maximum period to check for.
        min_period (int, optional): The minimum period to check for. Defaults to 1.
        mask_value (int, optional): The value to use to pad the array. Defaults to -1.
    """
    # make sure the mask_value is not in the array
    if (arr == mask_value).sum() > 0:
        raise ValueError("`mask_value` is in the array")

    # no since we can only detect sequences that repeat at least 3 times,
    # there is no point in checking for periods greater than 1/3 of the length
    max_period = min(max_period, len(arr) // 3)

    for period in range(min_period, max_period + 1):
        # pad the array so that it can be reshaped into a matrix matching the period
        padded_arr = np.pad(arr, (0, period - (len(arr) % period)), constant_values=mask_value)
        shaped_arr = padded_arr.reshape(-1, period)

        # find rows that are equal to the previous  row; these are the possibly-periodic sequences
        is_equal_to_prev_row = shaped_arr == np.roll(shaped_arr, shift=1, axis=0)
        rows_with_period, *_ = np.where(is_equal_to_prev_row.all(axis=1))

        # no sequences found with this period
        if len(rows_with_period) == 0:
            continue

        # this finds the start and end positions of the sequences with period `period`
        where_true_consecutive = group_consecutive_values(rows_with_period)

        for sequence in where_true_consecutive:
            start_row = sequence[0]
            end_row = sequence[-1]

            # we check if any value at the end of the previous row is True, e.g.:
            #     [[False, False, True, True]
            #      [True, True, True, True]]
            # (in the case above, start offset is 2). If so, we subtract that from the
            # period to get the actual start offset.
            start_offset = find_start_last_consecutive_true(is_equal_to_prev_row[start_row - 1])
            start_offset = period - start_offset if start_offset > 0 else 0

            # same idea as above, we want to compute offset. Only difference is that
            # `find_end_first_consecutive_true` already returns the offset, so we don't
            # need to subtract from the period.
            end_offset = find_end_first_consecutive_true(is_equal_to_prev_row[end_row + 1])

            # because we are always comparing with preceding row in
            # `is_equal_to_prev_row`, we need to subtract 1 from the row number
            start_pos = (start_row - 1) * period - start_offset

            # note that the end position is exclusive
            end_pos = ((end_row + 1) * period) + end_offset

            out = RepetitionTuple(
                start=start_pos, end=end_pos, period=period, times=(end_pos - start_pos) // period
            )
            if out.times > 2:
                # cannot accurately determine the period of a sequence that repeats
                # less than 3 times with this algorithm
                yield out



import os

import pandas as pd

PATH = "/data/niklas/dolma/stack"
TABLE = "Language & RPJ % flagged & SC % flagged & RPJ SC Corr."
TABLE += "& Language & RPJ % flagged & SC % flagged & RPJ SC Corr."
TABLE += "\\\\" + "\n"
TABLE_ROWS = []

paths = sorted(os.listdir(PATH))
halfway = len(paths) // 2

for i, lang in enumerate(paths):
    
    all_data = []

    for dataset in os.listdir(os.path.join(PATH, lang)):
        # Load json.gz
        data = pd.read_json(os.path.join(PATH, lang, dataset), lines=True)
        all_data.append(data)
    
    # Concatenate all dataframes
    all_data = pd.concat(all_data)
    # Cast to int
    all_data["rpj"] = all_data["rpj"].astype(int)
    all_data["starcoder"] = all_data["starcoder"].astype(int)
    rpj_flagged = round(100 * (all_data["rpj"].sum() / len(all_data)), 1)
    sc_flagged = round(100 * (all_data["starcoder"].sum() / len(all_data)), 1)
    if rpj_flagged == 0 or sc_flagged == 0:
        corr = "N/A"
    else:
        corr = round(all_data["rpj"].corr(all_data["starcoder"]), 3)

    if i >= halfway:
        TABLE_ROWS[i-halfway] += f" & {lang} & {rpj_flagged} & {sc_flagged} & {corr} \\\\"
    else:
        TABLE_ROWS.append(f"{lang} & {rpj_flagged} & {sc_flagged} & {corr}")
        # TABLE_ROWS.append(f"{lang} & {rpj_flagged} & {sc_flagged} & {corr} \\\\")

TABLE += "\n".join(TABLE_ROWS)
print(TABLE)



import logging
import multiprocessing


def get_logger(name: str) -> logging.Logger:
    if (proc_name := multiprocessing.current_process().name) == "MainProcess":
        proc_name = "main"
    proc_name = proc_name.replace(" ", "_")

    name = f"{proc_name}.dolma.{name}"
    logger = logging.getLogger(name)
    logger.setLevel(logging.WARN)

    if not logger.handlers:
        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter("%(asctime)s %(levelname)s %(name)s %(message)s"))
        logger.addHandler(handler)

    return logger



import unittest

import numpy as np

from dolma.core.binning import (
    FixedBucketsValTracker,
    InferBucketsValTracker,
    merge_bins,
)


class TestBinning(unittest.TestCase):
    def setUp(self) -> None:
        np.random.seed(0)

    def test_binning(self):
        bin_a = np.arange(0, 10_000, 55).astype(np.float64)
        bin_b = np.arange(0, 10_000, 100).astype(np.float64)

        count_a = np.random.randint(0, 15, len(bin_a))
        count_b = np.random.randint(0, 15, len(bin_b))

        bin_c, count_c = merge_bins(bin_a, count_a, bin_b, count_b)

        self.assertEqual(set(bin_c), set(np.concatenate([bin_a, bin_b])))
        self.assertEqual(sum(count_c), sum(count_a) + sum(count_b))

    def test_binning_with_repetitions(self):
        bin_a = np.random.randint(0, 100, 10_000).astype(np.float64)
        bin_b = np.random.randint(0, 100, 1_000).astype(np.float64)

        bin_a = np.sort(bin_a)
        bin_b = np.sort(bin_b)

        count_a = np.ones_like(bin_a, dtype=np.int64)
        count_b = np.ones_like(bin_b, dtype=np.int64)

        bin_c, count_c = merge_bins(bin_a, count_a, bin_b, count_b)

        self.assertEqual(set(bin_c), set(np.concatenate([bin_a, bin_b])))
        self.assertEqual(sum(count_c), sum(count_a) + sum(count_b))

    def test_binning_no_overlap(self):
        bin_a = np.arange(0, 1_000, 3).astype(np.float64)
        bin_b = np.arange(17, 1_000, 17).astype(np.float64)

        count_a = np.ones_like(bin_a, dtype=np.int64)
        count_b = np.ones_like(bin_b, dtype=np.int64)

        bin_c, count_c = merge_bins(bin_a, count_a, bin_b, count_b)

        self.assertEqual(set(bin_c), set(np.concatenate([bin_a, bin_b])))
        self.assertEqual(sum(count_c), sum(count_a) + sum(count_b))

    def test_bucket_val_trackers(self):
        tracker = InferBucketsValTracker(n=100_000)

        values = np.random.randn(1_000_000)

        for v in values:
            tracker.add(v)

        tracker_counts, tracker_bins = tracker.summarize(10)

        self.assertEqual(sum(tracker_counts), len(values))
        self.assertEqual(sorted(tracker_bins), tracker_bins)

        tracker_dist, tracker_bins = tracker.summarize(10, density=True)
        hist_dist, hist_bins = np.histogram(values, bins=10, density=True)

        for td, hd in zip(tracker_dist, hist_dist):
            self.assertAlmostEqual(np.abs(td - hd), 0, delta=0.1)

        for tb, hb in zip(tracker_bins, hist_bins):
            self.assertAlmostEqual(np.abs(tb - hb), 0, delta=0.5)


class FixedBinning(unittest.TestCase):
    def setUp(self) -> None:
        np.random.seed(0)

    def test_normal_bins(self):
        tr = FixedBucketsValTracker()
        vals = np.random.randn(2_000_000) * 100
        total_count = len(vals)

        for v in vals:
            tr.add(v)

        for (tr_c, tr_b), (hist_c, hist_b) in zip(zip(*tr.summarize(10)), zip(*np.histogram(vals, bins=10))):
            count_diff = np.abs(tr_c - hist_c) / total_count
            bin_diff = np.abs(tr_b - hist_b)
            self.assertLess(count_diff, 0.01)
            self.assertLess(bin_diff, 10)

    def test_uniform_bins(self):
        tr = FixedBucketsValTracker()
        vals = np.random.rand(2_000_000)
        total_count = len(vals)

        for v in vals:
            tr.add(v)

        for (tr_c, tr_b), (hist_c, hist_b) in zip(zip(*tr.summarize(10)), zip(*np.histogram(vals, bins=10))):
            count_diff = np.abs(tr_c - hist_c) / total_count
            bin_diff = np.abs(tr_b - hist_b)
            self.assertLess(count_diff, 0.01)
            self.assertLess(bin_diff, 0.01)



import hashlib
import os
from datetime import datetime
from multiprocessing import cpu_count
from tempfile import TemporaryDirectory
from typing import Any, Dict

import smart_open
from msgspec.json import Encoder
from necessary import necessary

from dolma.core.parallel import BaseParallelProcessor, QueueType

with necessary("pyarrow"):
    import pyarrow.parquet as pq


TIMESTAMP = datetime.now().strftime("%Y-%m-%dT%H:%M:%S.%f")[:23] + "Z"


class StarcoderFormatConvert(BaseParallelProcessor):

    @classmethod
    def increment_progressbar(  # type: ignore
        cls, queue: QueueType, /, files: int = 0, documents: int = 0
    ) -> Dict[str, int]:
        return super().increment_progressbar(queue, files=files, documents=documents)

    @classmethod
    def process_single(
        cls,
        source_path: str,
        destination_path: str,
        queue: QueueType,
        **kwargs: Any
    ):
        docs = 0
        print_every = 10_000
        enc = Encoder()

        destination_path = destination_path.replace(".parquet", ".jsonl.gz")

        *_, extension, __ = destination_path.split("/")

        with smart_open.open(destination_path, "wb") as output_file:
            table = pq.read_table(source_path)
            column_names = [str(c) for c in table.column_names]
            rows = (dict(zip(column_names, row)) for row in zip(*(table[c] for c in column_names)))
            for row in rows:
                text = str(row.pop('content', ''))
                if not text:
                    continue

                id_ = str(row.pop('id', hashlib.md5(text.encode()).hexdigest()))
                metadata = {key: str(value) for key, value in row.items()}
                metadata['extension'] = extension

                doc = {
                    'id': id_,
                    'text': text,
                    'source': 'starcoder',
                    'created': TIMESTAMP,
                    'added': TIMESTAMP,
                    'metadata': metadata
                }
                output_file.write(enc.encode(doc) + b'\n')  # type: ignore
                docs += 1

                if docs % print_every == 0:
                    cls.increment_progressbar(queue, documents=print_every)

        cls.increment_progressbar(queue, documents=docs % print_every, files=1)


def main():
    os.environ["PYTHONBREAKPOINT"] = "ipdb.set_trace"

    with TemporaryDirectory() as tempdir:
        src = "s3://ai2-llm/pretraining-data/sources/starcoder/raw/**/*.parquet"
        dst = "s3://ai2-llm/pretraining-data/sources/starcoder/v0/documents"

        converter = StarcoderFormatConvert(
            source_prefix=src,
            destination_prefix=dst,
            metadata_prefix=tempdir,
            num_processes=max(cpu_count() // 4, 1),
            debug=False
        )
        converter()


if __name__ == "__main__":
    main()



import itertools
import os
from pathlib import Path
from unittest import TestCase

from dolma.core.paths import (
    _escape_glob,
    _pathify,
    _unescape_glob,
    add_suffix,
    glob_path,
    is_glob,
    join_path,
    make_relative,
    split_glob,
    split_path,
    sub_prefix,
    sub_suffix,
)

from .utils import clean_test_data, get_test_prefix, skip_aws_tests, upload_s3_prefix

LOCAL_DATA = Path(__file__).parent.parent / "data"


class TestPaths(TestCase):
    def setUp(self) -> None:
        if skip_aws_tests():
            self.remote_test_prefix = None
        else:
            self.remote_test_prefix = get_test_prefix()
            upload_s3_prefix(s3_prefix=f"{self.remote_test_prefix}", local_prefix="tests/data/expected/*")

    def tearDown(self) -> None:
        if self.remote_test_prefix is not None:
            clean_test_data(self.remote_test_prefix)

    def test_pathify(self):
        path = "s3://path/to/file"
        protocol, path = _pathify(path)
        self.assertEqual(protocol, "s3")
        self.assertEqual(path, Path("path/to/file"))

        path = "path/to/file"
        protocol, path = _pathify(path)
        self.assertEqual(protocol, "")
        self.assertEqual(path, Path("path/to/file"))

        path = "/path/to/file"
        protocol, path = _pathify(path)
        self.assertEqual(protocol, "")
        self.assertEqual(path, Path("/path/to/file"))

    def test_local_glob_path(self):
        local_glob = str(LOCAL_DATA / "*.json.gz")
        paths = list(glob_path(local_glob))
        expected = [str(LOCAL_DATA / fn) for fn in os.listdir(LOCAL_DATA) if fn.endswith(".json.gz")]
        self.assertEqual(sorted(paths), sorted(expected))

    def test_remote_glob_path(self):
        if self.remote_test_prefix is None:
            return self.skipTest("Skipping AWS tests")

        paths = list(glob_path(f"{self.remote_test_prefix}/**/*.json.gz"))
        expected = [
            f"{self.remote_test_prefix}/tests/data/expected/{fn}"
            for fn in os.listdir(LOCAL_DATA / "expected")
            if fn.endswith(".json.gz")
        ]
        self.assertEqual(sorted(paths), sorted(expected))

    def test_local_glob_with_recursive(self):
        local_glob = str(LOCAL_DATA / "**/*-paragraphs.json.gz")
        paths = list(glob_path(local_glob))
        expected = list(
            itertools.chain.from_iterable(
                (str(fp),)
                if (fp := LOCAL_DATA / fn).is_file() and "paragraphs" in fn
                else ((str(fp / sn) for sn in os.listdir(fp) if "paragraphs" in sn) if fp.is_dir() else ())
                for fn in os.listdir(LOCAL_DATA)
            )
        )
        self.assertEqual(sorted(paths), sorted(expected))

    def test_sub_prefix(self):
        path_a = "s3://path/to/b/and/more"
        path_b = "s3://path/to/b"

        self.assertEqual(sub_prefix(path_a, path_b), "and/more")
        self.assertEqual(sub_prefix(path_b, path_a), path_b)

        path_c = "/path/to/c"
        path_d = "/path/to/c/and/more"

        self.assertEqual(sub_prefix(path_d, path_c), "and/more")
        self.assertEqual(sub_prefix(path_c, path_d), path_c)

        with self.assertRaises(ValueError):
            sub_prefix(path_a, path_c)

    def test_sub_suffix(self):
        path_a = "s3://path/to/dir/and/more"
        path_b = "and/more"
        self.assertEqual(sub_suffix(path_a, path_b), "s3://path/to/dir")

        path_c = "/path/to/dir/and/more"
        path_d = "path/to/dir/and/more"
        self.assertEqual(sub_suffix(path_c, path_d), "/")

    def test_add_prefix(self):
        path_a = "s3://path/to/b"
        path_b = "and/more"

        self.assertEqual(add_suffix(path_a, path_b), "s3://path/to/b/and/more")

        path_c = "/path/to/c"
        path_d = "and/more"

        self.assertEqual(add_suffix(path_c, path_d), "/path/to/c/and/more")

        with self.assertRaises(ValueError):
            add_suffix(path_a, path_c)
            add_suffix(path_c, path_a)
            add_suffix(path_a, path_a)

    def test_wildcard_operations(self):
        path_a = "s3://path/to/dir"
        self.assertEqual(add_suffix(path_a, "*"), "s3://path/to/dir/*")

        path_b = "/path/to/dir/**"
        self.assertEqual(add_suffix(path_b, "*"), "/path/to/dir/**/*")

        path_c = "s3://path/to/dir/**"
        self.assertEqual(sub_prefix(path_c, path_a), "**")

    def test_make_relative(self):
        paths = [
            "/path/to/dir/and/more",
            "/path/to/dir/and/**.zip",
            "/path/to/dir/more/**/stuff",
        ]
        base, relative_paths = make_relative(paths)
        self.assertEqual(base, "/path/to/dir")
        self.assertEqual(relative_paths, ["and/more", "and/**.zip", "more/**/stuff"])

        paths = [
            "/foo",
            "/bar/**",
            "/baz/**/**",
        ]
        base, relative_paths = make_relative(paths)
        self.assertEqual(base, "/")
        self.assertEqual(relative_paths, ["foo", "bar/**", "baz/**/**"])

        paths = [
            "s3://path/to/a/and/b",
            "s3://path/to/a/and/**.zip",
            "s3://path/to/b/more/**/stuff",
        ]

        base, relative_paths = make_relative(paths)
        self.assertEqual(base, "s3://path/to")
        self.assertEqual(relative_paths, ["a/and/b", "a/and/**.zip", "b/more/**/stuff"])

        paths = [
            "s3://path_a/to/dir/and/more",
            "s3://path_b/to/dir/and/**.zip",
            "s3://path_c/to/dir/more/**/stuff",
        ]

        base, relative_paths = make_relative(paths)
        self.assertEqual(base, "s3://")
        self.assertEqual(
            relative_paths,
            [
                "path_a/to/dir/and/more",
                "path_b/to/dir/and/**.zip",
                "path_c/to/dir/more/**/stuff",
            ],
        )

    def test_split(self):
        prot, parts = split_path("s3://path/to/dir")
        self.assertEqual(prot, "s3")
        self.assertEqual(parts, ("path", "to", "dir"))

        prot, parts = split_path("s3://path/to/[abc]/dir")
        self.assertEqual(prot, "s3")
        self.assertEqual(parts, ("path", "to", "[abc]", "dir"))

        prot, parts = split_path("/path/to/dir")
        self.assertEqual(prot, "")
        self.assertEqual(parts, ("/", "path", "to", "dir"))

        prot, parts = split_path("s3://path/to/dir/")
        self.assertEqual(prot, "s3")
        self.assertEqual(parts, ("path", "to", "dir"))

        prot, parts = split_path("path/to/dir/")
        self.assertEqual(prot, "")
        self.assertEqual(parts, ("path", "to", "dir"))

    def test_join(self):
        path = join_path("s3", ("path", "to", "dir"))
        self.assertEqual(path, "s3://path/to/dir")

        path = join_path("", ("/", "path", "to", "dir"))
        self.assertEqual(path, "/path/to/dir")

        path = join_path("s3", ("/", "path", "to", "dir", ""))
        self.assertEqual(path, "s3://path/to/dir")

        path = join_path("", ("path", "to", "dir", ""))
        self.assertEqual(path, "path/to/dir")

        path = join_path("s3", ("path/to", "dir/"))
        self.assertEqual(path, "s3://path/to/dir")

        path = join_path("", "path", ("to", "a", "very", "hidden"), "dir" "/with_file.txt")
        self.assertEqual(path, "path/to/a/very/hidden/dir/with_file.txt")

    def test_split_and_join(self):
        paths = [
            "s3://path/to/a/and/b",
            "s3://path/to/a/and/**.zip",
            "s3://path/to/b/more/**/stuff",
            "/path/to/dir/and/more",
            "/path/to/dir/and/**.zip",
            "*",
        ]
        for path in paths:
            self.assertEqual(join_path(*split_path(path)), path)

    def test_is_glob(self):
        path = "s3://path/to/a/and/b"
        self.assertFalse(is_glob(path))

        path = "/user/?/docs/*"
        self.assertTrue(is_glob(path))

        path = "/user/1/docs/file.txt"
        self.assertFalse(is_glob(path))

        path = r"/user/\?/docs/*"
        self.assertTrue(is_glob(path))

        path = r"/user/\?/docs/\*"
        self.assertFalse(is_glob(path))

    def test_escape_glob(self):
        path = "/user/?/docs/*"
        self.assertEqual(_escape_glob(path), "/user/\u2582/docs/\u2581")

        path = "/user/[abc]/docs/*"
        self.assertEqual(_escape_glob(path), "/user/\u2583abc\u2584/docs/\u2581")

        path = "/user/[abc]/docs/[a-z]/\\*"
        self.assertEqual(_escape_glob(path), "/user/\u2583abc\u2584/docs/\u2583a-z\u2584/\\*")

        path = "/no/glob/here"
        self.assertEqual(_escape_glob(path), path)

    def test_unescape_glob(self):
        path = "/user/?/docs/*"
        self.assertEqual(_unescape_glob(_escape_glob(path)), path)

        path = "/user/[abc]/docs/*"
        self.assertEqual(_unescape_glob(_escape_glob(path)), path)

        path = "/user/[abc]/docs/[a-z]/\\*"
        self.assertEqual(_unescape_glob(_escape_glob(path)), path)

        path = "/no/glob/here"
        self.assertEqual(_unescape_glob(_escape_glob(path)), path)

    def test_split_glob(self):
        path = "/user/_/docs/*"
        self.assertEqual(split_glob(path), ("/user/_/docs", "*"))

        path = "/user/?/docs/*"
        self.assertEqual(split_glob(path), ("/user", "?/docs/*"))

        path = "/user/[abc]/docs/*"
        self.assertEqual(split_glob(path), ("/user", "[abc]/docs/*"))

        path = "/user/\\[abc\\]/docs/[a-z]/\\*"
        self.assertEqual(split_glob(path), ("/user/\\[abc\\]/docs", "[a-z]/\\*"))

        path = "/no/glob/here"
        self.assertEqual(split_glob(path), ("/no/glob/here", ""))



import warnings

# warning raised by pkg_resources used in a lot of google packages
warnings.filterwarnings("ignore", message=r".*declare_namespace\(\'.*google.*", category=DeprecationWarning)
# base warning raised when warning above are raised
warnings.filterwarnings("ignore", message=r".*pkg_resources is deprecated.*", category=DeprecationWarning)



import json
from pathlib import Path
from tempfile import NamedTemporaryFile
from unittest import TestCase

from dolma.cli.__main__ import main

from .utils import (
    clean_test_data,
    download_s3_prefix,
    get_test_prefix,
    load_jsonl,
    skip_aws_tests,
    upload_s3_prefix,
)

EMAIL_SPANS = Path(__file__).parent.parent / "config/email-spans.json"
FILTER_BY_SPANS = Path(__file__).parent.parent / "config/filter-by-spans.json"
MIXER = Path(__file__).parent.parent / "config/mixer.json"
PARAGRAPH_SPANS = Path(__file__).parent.parent / "config/paragraph-spans.json"


class TestMixer(TestCase):
    def setUp(self) -> None:
        if skip_aws_tests():
            self.remote_test_prefix = None
        else:
            self.remote_test_prefix = get_test_prefix()
            upload_s3_prefix(s3_prefix=f"{self.remote_test_prefix}", local_prefix="tests/data/provided/**/*.gz")
            upload_s3_prefix(
                s3_prefix=f"{self.remote_test_prefix}", local_prefix="tests/data/provided/attributes/**/*.gz"
            )

    def tearDown(self) -> None:
        if self.remote_test_prefix is not None:
            clean_test_data(self.remote_test_prefix)

    def test_email_spans(self):
        main(argv=["-c", str(EMAIL_SPANS), "mix"])

        self.assertEqual(
            load_jsonl("tests/data/expected/email-spans.json.gz"),
            load_jsonl("tests/work/output/email-spans/email-spans-0000.json.gz"),
        )

    def test_filter_by_spans(self):
        main(argv=["-c", str(FILTER_BY_SPANS), "mix"])

        self.assertEqual(
            load_jsonl("tests/data/expected/filter-by-spans.json.gz"),
            load_jsonl("tests/work/output/filter-by-spans/filter-by-spans-test-0000.json.gz"),
        )

    def test_mixer(self):
        main(argv=["-c", str(MIXER), "mix"])

        self.assertEqual(
            load_jsonl("tests/data/expected/mixer.json.gz"),
            load_jsonl("tests/work/output/mixer/mixer-test-0000.json.gz"),
        )

    def test_paragraph_spans(self):
        main(argv=["-c", str(PARAGRAPH_SPANS), "mix"])

        self.assertEqual(
            load_jsonl("tests/data/expected/remove-paragraphs.json.gz"),
            load_jsonl("tests/work/output/paragraph-spans/paragraph-spans-test-0000.json.gz"),
        )

    def test_local_input_remote_output(self):
        if self.remote_test_prefix is None:
            return self.skipTest("Skipping AWS tests")

        with open(MIXER, mode="r", encoding="utf8") as f:
            config = json.load(f)

        # keep track of local output path
        local_output = config["streams"][0]["output"]["path"]

        # replace results path with s3 path
        config["streams"][0]["output"]["path"] = f"{self.remote_test_prefix}/{local_output}"

        with NamedTemporaryFile("w") as f:
            json.dump(config, f)
            f.flush()

            main(argv=["-c", f.name, "mix"])

        download_s3_prefix(f"{self.remote_test_prefix}/tests/work", "tests/work/remote")

        self.assertEqual(
            load_jsonl("tests/data/expected/mixer.json.gz"),
            load_jsonl("tests/work/remote/output/mixer/mixer-test-0000.json.gz"),
        )

    def test_remote_input_remote_output(self):
        if self.remote_test_prefix is None:
            return self.skipTest("Skipping AWS tests")

        with open(MIXER, mode="r", encoding="utf8") as f:
            config = json.load(f)

        # keep track of local output path
        local_input = config["streams"][0]["documents"][0]
        local_output = config["streams"][0]["output"]["path"]

        # replace results path with s3 path
        config["streams"][0]["output"]["path"] = f"{self.remote_test_prefix}/{local_output}"

        # upload local input to s3, replace local input with s3 path
        config["streams"][0]["documents"][0] = f"{self.remote_test_prefix}/{local_input}"

        with NamedTemporaryFile("w") as f:
            json.dump(config, f)
            f.flush()

            main(argv=["-c", f.name, "mix"])

        download_s3_prefix(f"{self.remote_test_prefix}/tests/work", "tests/work/remote")

        self.assertEqual(
            load_jsonl("tests/data/expected/mixer.json.gz"),
            load_jsonl("tests/work/remote/output/mixer/mixer-test-0000.json.gz"),
        )

    def test_remote_input_local_output(self):
        if self.remote_test_prefix is None:
            return self.skipTest("Skipping AWS tests")

        with open(MIXER, mode="r", encoding="utf8") as f:
            config = json.load(f)

        # keep track of local output path
        local_input = config["streams"][0]["documents"][0]

        # upload local input to s3, replace local input with s3 path
        config["streams"][0]["documents"][0] = f"{self.remote_test_prefix}/{local_input}"

        with NamedTemporaryFile("w") as f:
            json.dump(config, f)
            f.flush()

            main(argv=["-c", f.name, "mix"])

        self.assertEqual(
            load_jsonl("tests/data/expected/mixer.json.gz"),
            load_jsonl("tests/work/output/mixer/mixer-test-0000.json.gz"),
        )
